from rest_framework import mixins
from rest_framework.generics import GenericAPIView
from rest_framework.viewsets import ModelViewSet

from . import mixins as bulk_mixins

__all__ = [
    "BulkCreateAPIView",
    "BulkDestroyAPIView",
    "BulkModelViewSet",
    "BulkUpdateAPIView",
    "ListBulkCreateAPIView",
    "ListBulkCreateDestroyAPIView",
    "ListBulkCreateUpdateAPIView",
    "ListBulkCreateUpdateDestroyAPIView",
    "ListCreateBulkUpdateAPIView",
    "ListCreateBulkUpdateDestroyAPIView",
]


# ################################################## #
# Concrete view classes that provide method handlers #
# by composing the mixin classes with the base view. #
# ################################################## #


class BulkCreateAPIView(bulk_mixins.BulkCreateModelMixin, GenericAPIView):
    def post(self, request, *args, **kwargs):
        return self.create(request, *args, **kwargs)


class BulkUpdateAPIView(bulk_mixins.BulkUpdateModelMixin, GenericAPIView):
    def put(self, request, *args, **kwargs):
        return self.bulk_update(request, *args, **kwargs)

    def patch(self, request, *args, **kwargs):
        return self.partial_bulk_update(request, *args, **kwargs)


class BulkDestroyAPIView(bulk_mixins.BulkDestroyModelMixin, GenericAPIView):
    def delete(self, request, *args, **kwargs):
        return self.bulk_destroy(request, *args, **kwargs)


class ListBulkCreateAPIView(
    mixins.ListModelMixin, bulk_mixins.BulkCreateModelMixin, GenericAPIView
):
    def get(self, request, *args, **kwargs):
        return self.list(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        return self.create(request, *args, **kwargs)


class ListCreateBulkUpdateAPIView(
    mixins.ListModelMixin,
    mixins.CreateModelMixin,
    bulk_mixins.BulkUpdateModelMixin,
    GenericAPIView,
):
    def get(self, request, *args, **kwargs):
        return self.list(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        return self.create(request, *args, **kwargs)

    def put(self, request, *args, **kwargs):
        return self.bulk_update(request, *args, **kwargs)

    def patch(self, request, *args, **kwargs):
        return self.partial_bulk_update(request, *args, **kwargs)


class ListCreateBulkUpdateDestroyAPIView(
    mixins.ListModelMixin,
    mixins.CreateModelMixin,
    bulk_mixins.BulkUpdateModelMixin,
    bulk_mixins.BulkDestroyModelMixin,
    GenericAPIView,
):
    def get(self, request, *args, **kwargs):
        return self.list(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        return self.create(request, *args, **kwargs)

    def put(self, request, *args, **kwargs):
        return self.bulk_update(request, *args, **kwargs)

    def patch(self, request, *args, **kwargs):
        return self.partial_bulk_update(request, *args, **kwargs)

    def delete(self, request, *args, **kwargs):
        return self.bulk_destroy(request, *args, **kwargs)


class ListBulkCreateUpdateAPIView(
    mixins.ListModelMixin,
    bulk_mixins.BulkCreateModelMixin,
    bulk_mixins.BulkUpdateModelMixin,
    GenericAPIView,
):
    def get(self, request, *args, **kwargs):
        return self.list(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        return self.create(request, *args, **kwargs)

    def put(self, request, *args, **kwargs):
        return self.bulk_update(request, *args, **kwargs)

    def patch(self, request, *args, **kwargs):
        return self.partial_bulk_update(request, *args, **kwargs)


class ListBulkCreateDestroyAPIView(
    mixins.ListModelMixin,
    bulk_mixins.BulkCreateModelMixin,
    bulk_mixins.BulkDestroyModelMixin,
    GenericAPIView,
):
    def get(self, request, *args, **kwargs):
        return self.list(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        return self.create(request, *args, **kwargs)

    def delete(self, request, *args, **kwargs):
        return self.bulk_destroy(request, *args, **kwargs)


class ListBulkCreateUpdateDestroyAPIView(
    mixins.ListModelMixin,
    bulk_mixins.BulkCreateModelMixin,
    bulk_mixins.BulkUpdateModelMixin,
    bulk_mixins.BulkDestroyModelMixin,
    GenericAPIView,
):
    def get(self, request, *args, **kwargs):
        return self.list(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        return self.create(request, *args, **kwargs)

    def put(self, request, *args, **kwargs):
        return self.bulk_update(request, *args, **kwargs)

    def patch(self, request, *args, **kwargs):
        return self.partial_bulk_update(request, *args, **kwargs)

    def delete(self, request, *args, **kwargs):
        return self.bulk_destroy(request, *args, **kwargs)


# ########################################################## #
# Concrete viewset classes that provide method handlers      #
# by composing the bulk mixin classes with the base viewset. #
# ########################################################## #


class BulkModelViewSet(
    bulk_mixins.BulkCreateModelMixin,
    bulk_mixins.BulkUpdateModelMixin,
    bulk_mixins.BulkDestroyModelMixin,
    ModelViewSet,
):
    pass



from rest_framework.renderers import BrowsableAPIRenderer


class FilterlessBrowsableAPIRenderer(BrowsableAPIRenderer):
    """Custom browsable API renderer that doesn't show filter forms."""

    def get_filter_form(self, data, view, request):
        """
        Disable filter form display.

        This is because of major performance problems with large installations,
        especially with large sets of related objects.

        FIXME(jathan): Revisit this after browsable API rendering has improved
        in future versions of DRF.
        """
        return



"""
This configuration file is just Python code. You may override any global
defaults by specifying them here.

For more information on this file, see
https://docs.djangoproject.com/en/1.8/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.8/ref/settings/
"""

SECRET_KEY = "fMK68NKgazLCjjTXjDtthhoRUS8IV4lwD-9G7iVd2Xs="
import os.path

from nsot.conf.settings import *

# Path where the config is found.
CONF_ROOT = os.path.dirname(__file__)

# A boolean that turns on/off debug mode. Never deploy a site into production
# with DEBUG turned on.
# Default: False
DEBUG = False

############
# Database #
############
# https://docs.djangoproject.com/en/dev/ref/settings/#databases
DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": os.path.join(CONF_ROOT, "nsot.sqlite3"),
        "USER": "nsot",
        "PASSWORD": "",
        "HOST": "",
        "PORT": "",
    }
}

###############
# Application #
###############

# The address on which the application will listen.
# Default: localhost
NSOT_HOST = "localhost"

# The port on which the application will be accessed.
# Default: 8990
NSOT_PORT = 8990

# The number of gunicorn worker processes for handling requests.
# Default: 4
NSOT_NUM_WORKERS = 4

# Timeout in seconds before gunicorn workers are killed/restarted.
# Default: 30
NSOT_WORKER_TIMEOUT = 30

# If True, serve static files directly from the app.
# Default: True
SERVE_STATIC_FILES = True

############
# Security #
############

# A URL-safe base64-encoded 32-byte key. This must be kept secret. Anyone with
# this key is able to create and read messages. This key is used for
# encryption/decryption of sessions and auth tokens. A unique key is randomly
# generated for you when you utilize ``nsot-server init``
# https://cryptography.io/en/latest/fernet/#cryptography.fernet.Fernet.generate_key

# Header to check for Authenticated Email. This is intended for use behind an
# authenticating reverse proxy.
USER_AUTH_HEADER = "X-NSoT-Email"

# The age, in seconds, until an AuthToken granted by the API will expire.
# Default: 600
AUTH_TOKEN_EXPIRY = 600  # 10 minutes

# A list of strings representing the host/domain names that this Django site can
# serve. This is a security measure to prevent an attacker from poisoning caches
# and triggering password reset emails with links to malicious hosts by
# submitting requests with a fake HTTP Host header, which is possible even under
# many seemingly-safe web server configurations.
# https://docs.djangoproject.com/en/1.8/ref/settings/#allowed-hosts
ALLOWED_HOSTS = ["*"]

##############
# Interfaces #
##############

# The default format for displaying MAC addresses. This defaults to
# ":"-separated and expanded (e.g. '00:00:00:00:00:00')
MACADDRESS_DEFAULT_DIALECT = "macaddress.mac_linux"

# The default speed in Mbps for newly device interfaces if not otherwise
# specified.
INTERFACE_DEFAULT_SPEED = 1000  # In Mbps (e.g. 1Gbps)

# Whether to compress IPv6 for display purposes, for example:
# - Default: 2620:0100:6000:0000:0000:0000:0000:0000/40
# - Compressed: 2620:100:6000::/40
# Default: True
NSOT_COMPRESS_IPV6 = True

# Temp debug logging
if os.getenv("NSOT_DEBUG"):
    DEBUG = True
    LOGGING["loggers"]["nsot"]["level"] = "DEBUG"
    LOGGING["loggers"]["django.db.backends"] = {
        "handlers": ["console"],
        "level": "DEBUG",
    }



# -*- coding: utf-8 -*-

from django.db import migrations, models
import django_extensions.db.fields.json

class Migration(migrations.Migration):

    dependencies = [
        ("nsot", "0021_remove_resource_object"),
    ]

    operations = [
        migrations.AddField(
            model_name="interface",
            name="_addresses_cache",
            field=django_extensions.db.fields.json.JSONField(
                default="[]", blank=True
            ),
        ),
        migrations.AddField(
            model_name="interface",
            name="_networks_cache",
            field=django_extensions.db.fields.json.JSONField(
                default="[]", blank=True
            ),
        ),
    ]



# Generated by Django 5.2.11 on 2026-02-12 08:32

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('nsot', '0040_index_together_to_indexes'),
    ]

    operations = [
        migrations.AddField(
            model_name='interface',
            name='mtu',
            field=models.IntegerField(blank=True, db_index=True, default=None, help_text='MTU (Maximum Transmission Unit) in bytes.', null=True),
        ),
    ]



from django.db import models

from .. import validators


class Site(models.Model):
    """A namespace for attribtues, devices, and networks."""

    name = models.CharField(
        max_length=255, unique=True, help_text="The name of the Site."
    )
    description = models.TextField(
        default="", blank=True, help_text="A helpful description for the Site."
    )

    def __str__(self):
        return self.name

    def clean_name(self, value):
        return validators.validate_name(value)

    def clean_fields(self, exclude=None):
        self.name = self.clean_name(self.name)

    def save(self, *args, **kwargs):
        self.full_clean()  # First validate fields are correct
        super().save(*args, **kwargs)

    def to_dict(self):
        return {
            "id": self.id,
            "name": self.name,
            "description": self.description,
        }



# -*- coding: utf-8 -*-

from django.db import migrations, models

from nsot.util import slugify_interface

def add_name_slug(apps, schema_editor):
    """Correctly name_slug for every Interface with slash in the name."""
    Interface = apps.get_model("nsot", "Interface")
    for i in Interface.objects.iterator():
        name_slug = slugify_interface(device_hostname=i.device_hostname, name=i.name)
        i.name_slug = name_slug
        i.save()

def remove_name_slug(apps, schema_editor):
    Interface = apps.get_model("nsot", "Interface")
    Interface.objects.update(name_slug=None)

class Migration(migrations.Migration):

    dependencies = [
        ("nsot", "0034_populate_interface_name_slug"),
    ]

    operations = [migrations.RunPython(add_name_slug, remove_name_slug)]



from django.contrib import admin
from django.contrib.auth import get_user_model
from django.contrib.auth.admin import UserAdmin as DjangoUserAdmin
from django.utils.translation import gettext_lazy as _
from guardian.admin import GuardedModelAdmin

from . import models


# Register our custom User model
class UserAdmin(DjangoUserAdmin):
    fieldsets = (
        (
            None,
            {
                "fields": ("email", "secret_key", "password"),
            },
        ),
        (
            _("Permissions"),
            {
                "fields": (
                    "is_active",
                    "is_staff",
                    "is_superuser",
                    "groups",
                    "user_permissions",
                ),
            },
        ),
        (_("Important dates"), {"fields": ("last_login", "date_joined")}),
    )
    add_fieldsets = (
        (
            None,
            {
                "classes": ("wide",),
                "fields": ("email", "password1", "password2"),
            },
        ),
    )
    list_display = ("email", "is_staff")
    list_filter = ("is_staff", "is_superuser", "is_active", "groups")
    search_fields = ("email",)
    ordering = ("email",)
    filter_horizontal = (
        "groups",
        "user_permissions",
    )


admin.site.register(get_user_model(), UserAdmin)


class SiteAdmin(GuardedModelAdmin):
    list_display = ("name", "description")
    list_filter = ("name",)


admin.site.register(models.Site, SiteAdmin)


class AttributeAdmin(GuardedModelAdmin):
    list_display = (
        "name",
        "resource_name",
        "description",
        "required",
        "display",
        "multi",
        "site",
    )
    list_filter = ("name", "resource_name", "required", "multi", "site")


admin.site.register(models.Attribute, AttributeAdmin)


class ValueAdmin(GuardedModelAdmin):
    list_display = ("name", "value", "resource_name", "resource_id")
    list_filter = ("name", "value", "resource_name")


admin.site.register(models.Value, ValueAdmin)


class ChangeAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "user",
        "event",
        "resource_name",
        "resource_id",
        "get_change_at",
        "resource_name",
        "site",
    )
    list_filter = ("event", "site")


admin.site.register(models.Change, ChangeAdmin)


class DeviceAdmin(GuardedModelAdmin):
    list_display = ("hostname", "site")
    list_filter = ("site",)

    fields = list_display


admin.site.register(models.Device, DeviceAdmin)


class NetworkAdmin(GuardedModelAdmin):
    mptt_level_indent = 10
    mptt_indent_field = "cidr"
    list_display = (
        "cidr",
        "network_address",
        "prefix_length",
        "ip_version",
        "is_ip",
        "parent",
        "site",
    )
    list_filter = ("prefix_length", "is_ip", "ip_version", "site")

    def get_cidr(self):
        return "%s/%s" % (self.network_address, self.prefix_length)

    fields = (
        "network_address",
        "broadcast_address",
        "prefix_length",
        "ip_version",
        "is_ip",
        "site",
    )


admin.site.register(models.Network, NetworkAdmin)


class InterfaceAdmin(GuardedModelAdmin):
    list_display = ("name", "device", "parent", "mac_address", "type", "speed")
    list_filter = ("type", "speed")

    fields = list_display


admin.site.register(models.Interface, InterfaceAdmin)


class ProtocolTypeAdmin(GuardedModelAdmin):
    list_display = ("name", "description", "site")
    list_filter = ("name", "site")


admin.site.register(models.ProtocolType, ProtocolTypeAdmin)


class ProtocolAdmin(GuardedModelAdmin):
    list_display = ("type", "description", "device", "interface", "circuit")
    list_filter = ("type", "description", "device", "site")

    fields = (
        "type",
        "device",
        "interface",
        "circuit",
        "auth_string",
        "description",
        "site",
    )


admin.site.register(models.Protocol, ProtocolAdmin)



import logging
from datetime import timedelta

from django.db import models
from django.db.models.query_utils import Q
from django.utils import timezone

from .. import exc, fields, util
from .attribute import Attribute
from .value import Value

log = logging.getLogger(__name__)


class ResourceSetTheoryQuerySet(models.query.QuerySet):
    """
    Set theory QuerySet for Resource objects to add ``.set_query()`` method.

    For example::

        >>> qs = Network.objects.filter(network_address=u'10.0.0.0')
        >>> qs.set_query('owner=jathan +metro=lax')

    You may also search using regex by appending ``_regex`` to an attribtue
    name and providing a regex pattern as the value::

        >>> qs = Device.objects.set_query('role_regex=[bd]r')

    Which is functionally equivalent to::

        >>> qs = Device.objects.set_query('role=br +role=dr')
    """

    def set_query(self, query, site_id=None, unique=False):
        """
        Filter objects by set theory attribute-value ``query`` patterns.
        """
        objects = self
        if site_id is not None:
            objects = objects.filter(site=site_id)

        try:
            attributes = util.parse_set_query(query)
        except (ValueError, TypeError) as err:
            raise exc.ValidationError({"query": str(err)})

        resource_name = self.model.__name__

        # If there aren't any parsed attributes, don't return anything.
        if not attributes:
            if unique:
                raise exc.ValidationError(
                    {
                        "query": "Query empty, unable to provide %s"
                        % resource_name
                    }
                )
            return objects.none()

        # Iterate a/v pairs and combine query results using MySQL-compatible
        # set operations w/ the ORM
        log.debug("QUERY [start]: objects = %r", objects)
        for action, name, value in attributes:
            # Is this a regex pattern?
            regex_query = False
            if name.endswith("_regex"):
                name = name.replace("_regex", "")  # Keep attribute name
                regex_query = True
                log.debug("Regex enabled for %r" % name)

            # Attribute lookup params
            params = dict(name=name, resource_name=resource_name)
            # Only include site_id if it's set
            if site_id is not None:
                params["site_id"] = site_id

            # If an Attribute doesn't exist, the set query is invalid. Return
            # an empty queryset. (fix #99)
            try:
                attr = Attribute.objects.get(**params)
            except Attribute.DoesNotExist as err:
                raise exc.ValidationError(
                    {"query": "%s: %r" % (str(err).rstrip("."), name)}
                )

            # Set lookup params
            next_set_params = {
                "name": attr.name,
                "value": value,
                "resource_name": resource_name,
            }

            # If it's a regex query, swap ``value`` with ``value__regex``.
            if regex_query:
                next_set_params["value__regex"] = next_set_params.pop("value")

            next_set = Q(
                id__in=Value.objects.filter(**next_set_params).values_list(
                    "resource_id", flat=True
                )
            )

            # This is the MySQL-compatible manual implementation of set theory,
            # baby!
            if action == "union":
                log.debug("SQL UNION")
                objects = objects | self.filter(next_set)
            elif action == "difference":
                log.debug("SQL DIFFERENCE")
                objects = objects.exclude(next_set)
            elif action == "intersection":
                log.debug("SQL INTERSECTION")
                objects = objects.filter(next_set)
            else:
                raise exc.BadRequest("BAD SET QUERY: %r" % (action,))
            log.debug("QUERY [iter]: objects = %r", objects)

        count = objects.count()
        if unique and count != 1:
            # There can be only one
            msg = "Query returned %r results, but exactly 1 expected" % count
            raise exc.ValidationError({"query": msg})
        # Gotta call .distinct() or we might get dupes.
        return objects.distinct()

    def expired(self, expired=True):
        """Filter by expiration status.

        :param expired:
            If ``True`` (default), return resources whose ``expires_at`` is in
            the past. If ``False``, return resources that are **not** expired
            (either no expiry set, or expiry in the future).
        """
        now = timezone.now()
        if expired:
            return self.filter(expires_at__lte=now)
        return self.filter(Q(expires_at__isnull=True) | Q(expires_at__gt=now))

    def expiring_within_days(self, days=7):
        """Resources that expire within the next *days* days (not already expired).

        :param days:
            Number of days from now to look ahead. Defaults to ``7``.
        """
        now = timezone.now()
        cutoff = now + timedelta(days=days)
        return self.filter(expires_at__gt=now, expires_at__lte=cutoff)

    def by_attribute(self, name, value, site_id=None):
        """
        Lookup objects by Attribute ``name`` and ``value``.
        """
        resource_name = self.model._meta.model_name.title()
        query = self.filter(
            id__in=Value.objects.filter(
                name=name, value=value, resource_name=resource_name
            ).values_list("resource_id", flat=True)
        )

        if site_id is not None:
            query = query.filter(site=site_id)

        return query


class ResourceManager(models.Manager):
    """
    Manager for Resource objects that adds a special resource methods:

    + ``.set_query()`` - For performing set theory lookups by attribute-value
      string patterns
    + ``.by_attribute()`` - For looking up objects by attribute name/value.

    """

    queryset_class = ResourceSetTheoryQuerySet

    def get_queryset(self):
        return self.queryset_class(self.model, using=self._db)

    def set_query(self, query, site_id=None, unique=False):
        """
        Filter objects by set theory attribute-value string patterns.

        For example::

            >>> Network.objects.set_query('owner=jathan +cluster=sjc'}
            [<Device: foo-bar1>, <Device: foo-bar3>, <Device: foo-bar4>]

            >>> Network.objects.set_query('owner=gary -cluster=sjc'}
            [<Device: foo-bar2>]

            >>> Network.objects.set_query('owner=jathan foo=baz', unique=True}
            [<Device: foo-bar3>]

        :param query:
            Set theory query pattern

        :param site_id:
            ID of Site to filter results

        :param unique:
            Find exactly one match, error otherwise
        """
        return self.get_queryset().set_query(query, site_id, unique)

    def by_attribute(self, name, value, site_id=None):
        """
        Filter objects by Attribute ``name`` and ``value``.

        For example::

            >>> Interface.objects.by_attribute(name='vlan', value=300)
            [<Interface: device=1, name=eth0>]

        :param name:
            Attribute name

        :param value:
            Attribute value

        :param site_id:
            ID of Site to filter results
        """
        return self.get_queryset().by_attribute(name, value, site_id)

    def expired(self, expired=True):
        """Proxy to queryset ``expired()`` method."""
        return self.get_queryset().expired(expired=expired)

    def expiring_within_days(self, days=7):
        """Proxy to queryset ``expiring_within_days()`` method."""
        return self.get_queryset().expiring_within_days(days=days)


class Resource(models.Model):
    """Base for heirarchial Resource objects that may have attributes."""

    _attributes_cache = fields.JSONField(
        null=False,
        blank=True,
        help_text="Local cache of attributes. (Internal use only)",
    )

    expires_at = models.DateTimeField(
        null=True,
        blank=True,
        default=None,
        db_index=True,
        help_text=(
            "Optional expiration timestamp. Resources with expires_at in "
            "the past are considered expired."
        ),
    )

    # Subclasses set this to the field name to derive site from (e.g., 'device')
    site_source_field = None

    # If True, clean_site() returns Site object; if False, returns site_id
    site_returns_object = False

    # Whether to validate uniqueness in full_clean() during save()
    _validate_unique_on_save = True

    def __init__(self, *args, **kwargs):
        self._set_attributes = kwargs.pop("attributes", None)
        super().__init__(*args, **kwargs)

    class Meta:
        abstract = True

    # Implement .objects.set_query()
    objects = ResourceManager()

    @property
    def attributes(self):
        return Value.objects.filter(
            resource_name=self._resource_name, resource_id=self.id
        )

    @property
    def _resource_name(self):
        return self.__class__.__name__

    def _purge_attribute_index(self):
        self.attributes.all().delete()

    def get_attributes(self):
        """Return the JSON-encoded attributes as a dict."""
        return self._attributes_cache

    def set_attributes(self, attributes, valid_attributes=None, partial=False):
        """Validate and store the attributes dict as a JSON-encoded string.

        When ``partial=True`` (i.e. PATCH), only the provided attributes are
        merged on top of existing ones. Setting a value to ``None`` (JSON
        ``null``) deletes that attribute. Required-attribute validation runs
        against the final merged result to prevent deletion of required keys.
        """
        log.debug("Resource.set_attributes() attributes = %r", attributes)

        # If no attributes and it's a partial update, NOOP.
        if attributes is None and partial:
            return

        if not isinstance(attributes, dict):
            raise exc.ValidationError(
                {
                    "attributes": f"Expected dictionary but received {type(attributes)}"
                }
            )

        # A dict of valid Attribute objects for this resource, keyed by
        # attribute name. If not provided, defaults to all matching
        # resource_name.
        if valid_attributes is None:
            valid_attributes = Attribute.all_by_name(
                self._resource_name, self.site
            )
        log.debug(
            "Resource.set_attributes() valid_attributes = %r", valid_attributes
        )

        # For partial updates, merge incoming attributes with existing ones.
        # Note: concurrent PATCHes on the same resource could race (read-
        # merge-write). This is acceptable for SQLite (serialized writes)
        # and typical single-user IPAM workloads. For high-concurrency
        # deployments, callers should wrap in select_for_update().
        if partial:
            existing = self.get_attributes() or {}

            # Separate deletions (value is None) from updates.
            deletions = {k for k, v in attributes.items() if v is None}
            updates = {k: v for k, v in attributes.items() if v is not None}

            # Validate that all referenced attribute names exist.
            for name in list(deletions) + list(updates.keys()):
                if name not in valid_attributes:
                    msg = f"Attribute name ({name}) does not exist."
                    raise exc.ValidationError({"attributes": msg})

            # Build merged attribute dict: start with existing, apply updates,
            # remove deletions.
            merged = dict(existing)
            merged.update(updates)
            for key in deletions:
                merged.pop(key, None)

            # Use merged as the canonical attributes for the rest of the flow.
            attributes = merged
        else:
            deletions = set()

        # Apply default values for attributes that have defaults and aren't
        # provided. Skip attributes that were explicitly deleted via null in a
        # partial update — explicit deletion takes priority over defaults.
        for attr_name, attr in valid_attributes.items():
            if (
                attr_name not in attributes
                and attr_name not in deletions
                and attr.default is not None
            ):
                attributes[attr_name] = attr.default

        # Attributes that are required according to ``valid_attributes``, but
        # are not found in ``attributes``. For partial updates, this validates
        # the final merged result still has all required attributes (preventing
        # deletion of required attributes via null).
        missing_attributes = {
            attribute.name
            for attribute in valid_attributes.values()
            if attribute.required and attribute.name not in attributes
        }
        log.debug(
            "Resource.set_attributes() missing_attributes = %r",
            missing_attributes,
        )

        # It's an error to have any missing attributes
        if missing_attributes:
            names = ", ".join(missing_attributes)
            raise exc.ValidationError(
                {"attributes": f"Missing required attributes: {names}"}
            )

        # Check attribute dependencies (transitive)
        for attr_name, attr in valid_attributes.items():
            if attr_name not in attributes:
                continue
            all_deps = attr.get_all_dependencies()
            missing_deps = [
                dep.name for dep in all_deps if dep.name not in attributes
            ]
            if missing_deps:
                raise exc.ValidationError(
                    {
                        "attributes": f"Attribute '{attr_name}' requires: "
                        f"{', '.join(sorted(missing_deps))}"
                    }
                )

        # Run validation each attribute value and prepare them for DB
        # insertion, raising any validation errors immediately.
        inserts = []
        for name, value in attributes.items():
            if name not in valid_attributes:
                msg = f"Attribute name ({name}) does not exist."
                raise exc.ValidationError({"attributes": msg})

            if not isinstance(name, str):
                raise exc.ValidationError(
                    {"attributes": "Attribute names must be a string type."}
                )

            attribute = valid_attributes[name]
            inserts.extend(attribute.validate_value(value))

        # Purge all of our previously existing attribute values and recreate
        # them anew.
        # FIXME(jathan): This isn't exactly efficient. How can make gud?
        self._purge_attribute_index()
        for insert in inserts:
            Value.objects.create(
                obj=self,
                attribute_id=insert["attribute_id"],
                value=insert["value"],
            )

        self.clean_attributes()

    def clean_attributes(self):
        """Make sure that attributes are saved as JSON."""
        attrs = {}

        # Only fetch the fields we need.
        for a in self.attributes.only("name", "value", "attribute").iterator():
            if a.attribute.multi:
                if a.name not in attrs:
                    attrs[a.name] = []
                attrs[a.name].append(a.value)
            else:
                attrs[a.name] = a.value
        self._attributes_cache = attrs  # Cache the attributes

        return attrs

    def clean_site(self, value):
        """
        Ensure site is set. If not provided, derive from site_source_field.

        Subclasses set site_source_field to the name of the relation field
        to derive site from (e.g., 'device', 'endpoint_a'). Set
        site_returns_object=True to return the Site object instead of site_id.
        """
        # If value is provided, return as-is (use identity check to preserve
        # original behavior — only derive site when value is explicitly None)
        if value is not None:
            return value

        # If no source field configured, return value as-is
        if self.site_source_field is None:
            return value

        # Try to get the source object
        source = None
        try:
            source = getattr(self, self.site_source_field)
        except self._meta.get_field(
            self.site_source_field
        ).related_model.DoesNotExist:
            # Source relation not loaded; try via FK id
            fk_id = getattr(self, f"{self.site_source_field}_id", None)
            if fk_id is not None:
                field = self._meta.get_field(self.site_source_field)
                source = field.related_model.objects.get(id=fk_id)

        if source is None:
            return value

        # Return either site object or site_id based on class configuration
        if self.site_returns_object:
            site_value = source.site
            if not site_value:
                msg = (
                    "No site was provided and the provided Device does not "
                    "have a site defined"
                )
                raise exc.ValidationError({"site": msg})
            return site_value
        return source.site_id

    def save(self, *args, **kwargs):
        self._is_new = self.id is None  # Check if this is a new object.

        # Use incoming valid_attributes if they are provided.
        valid_attributes = kwargs.pop("valid_attributes", None)

        # Allow subclasses to skip full_clean (e.g., Network does its own)
        skip_full_clean = kwargs.pop("_skip_full_clean", False)
        if not skip_full_clean:
            self.full_clean(validate_unique=self._validate_unique_on_save)

        super().save(*args, **kwargs)

        # This is so that we can set the attributes on create/update, but if
        # the object is new, make sure that it doesn't persist if attributes
        # fail.
        if self._set_attributes is not None:
            try:
                # And set the attributes (if any)
                self.set_attributes(
                    self._set_attributes, valid_attributes=valid_attributes
                )
            except exc.ValidationError:
                # If attributes fail validation, and I'm a new object, delete
                # myself and re-raise the error.
                if self._is_new:
                    self.delete()
                raise



# -*- coding: utf-8 -*-

from django.db import migrations, models

from nsot.util import slugify

def add_name_slug(apps, schema_editor):
    """ Add a name_slug for every Circuit that doesn't already have one """

    Circuit = apps.get_model("nsot", "Circuit")
    for c in Circuit.objects.all():
        if not c.name_slug:
            c.name_slug = slugify(c.name)
            c.save()

def remove_name_slug(apps, schema_editor):
    Circuit = apps.get_model("nsot", "Circuit")
    for c in Circuit.objects.all():
        c.name_slug = None
        c.save()

class Migration(migrations.Migration):

    dependencies = [
        ("nsot", "0030_add_circuit_name_slug"),
    ]

    operations = [migrations.RunPython(add_name_slug, remove_name_slug)]



# -*- coding: utf-8 -*-

from django.db import migrations, models
import django.db.models.deletion
import django_extensions.db.fields.json

class Migration(migrations.Migration):

    dependencies = [
        ("nsot", "0028_populate_interface_device_hostname"),
    ]

    operations = [
        migrations.CreateModel(
            name="Circuit",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                (
                    "_attributes_cache",
                    django_extensions.db.fields.json.JSONField(
                        help_text="Local cache of attributes. (Internal use only)",
                        blank=True,
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        default="",
                        help_text="Unique display name of the Circuit. If not provided, defaults to '{device_a}:{interface_a}_{device_z}:{interface_z}'",
                        unique=True,
                        max_length=255,
                    ),
                ),
                (
                    "endpoint_a",
                    models.OneToOneField(
                        related_name="circuit_a",
                        on_delete=django.db.models.deletion.PROTECT,
                        verbose_name="A-side endpoint Interface",
                        to="nsot.Interface",
                        help_text="Unique ID of Interface at the A-side.",
                    ),
                ),
                (
                    "site",
                    models.ForeignKey(
                        related_name="circuits",
                        on_delete=django.db.models.deletion.PROTECT,
                        to="nsot.Site",
                        help_text="Unique ID of the Site this Circuit is under.",
                    ),
                ),
                (
                    "endpoint_z",
                    models.OneToOneField(
                        related_name="circuit_z",
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="nsot.Interface",
                        help_text="Unique ID of Interface at the Z-side.",
                        verbose_name="Z-side endpoint Interface",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.AlterField(
            model_name="attribute",
            name="resource_name",
            field=models.CharField(
                help_text="The name of the Resource to which this Attribute is bound.",
                max_length=20,
                verbose_name="Resource Name",
                db_index=True,
                choices=[
                    ("Device", "Device"),
                    ("Interface", "Interface"),
                    ("Network", "Network"),
                    ("Circuit", "Circuit"),
                ],
            ),
        ),
        migrations.AlterField(
            model_name="change",
            name="resource_name",
            field=models.CharField(
                help_text="The name of the Resource for this Change.",
                max_length=20,
                verbose_name="Resource Type",
                db_index=True,
                choices=[
                    ("Network", "Network"),
                    ("Attribute", "Attribute"),
                    ("Site", "Site"),
                    ("Interface", "Interface"),
                    ("Circuit", "Circuit"),
                    ("Device", "Device"),
                ],
            ),
        ),
        migrations.AlterField(
            model_name="value",
            name="resource_name",
            field=models.CharField(
                help_text="The name of the Resource type to which the Value is bound.",
                max_length=20,
                verbose_name="Resource Type",
                db_index=True,
                choices=[
                    ("Network", "Network"),
                    ("Attribute", "Attribute"),
                    ("Site", "Site"),
                    ("Interface", "Interface"),
                    ("Circuit", "Circuit"),
                    ("Device", "Device"),
                ],
            ),
        ),
    ]



# -*- coding: utf-8 -*-

from django.db import models, migrations
import django.db.models.deletion
import macaddress.fields

class Migration(migrations.Migration):

    dependencies = [
        ("nsot", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="Assignment",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("address", models.ForeignKey(
                    to="nsot.Network",
                    on_delete=django.db.models.deletion.CASCADE
                )),
            ],
        ),
        migrations.CreateModel(
            name="Interface",
            fields=[
                (
                    "resource_ptr",
                    models.OneToOneField(
                        parent_link=True,
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        to="nsot.Resource",
                        on_delete=django.db.models.deletion.CASCADE,
                    ),
                ),
                ("name", models.CharField(max_length=255, db_index=True)),
                ("description", models.CharField(default="", max_length=255)),
                (
                    "type",
                    models.IntegerField(
                        db_index=True,
                        verbose_name="Interface Type",
                        choices=[
                            ("other", 1),
                            ("ethernet", 6),
                            ("loopback", 24),
                            ("tunnel", 131),
                            ("l2vlan", 135),
                            ("l3vlan", 136),
                            ("mpls", 150),
                            ("lag", 161),
                        ],
                    ),
                ),
                (
                    "mac",
                    macaddress.fields.MACAddressField(
                        integer=True, db_index=True, blank=True
                    ),
                ),
                ("speed", models.IntegerField(db_index=True)),
                (
                    "addresses",
                    models.ManyToManyField(
                        related_name="addresses",
                        through="nsot.Assignment",
                        to="nsot.Network",
                        db_index=True,
                    ),
                ),
                (
                    "device",
                    models.ForeignKey(
                        related_name="interfaces",
                        to="nsot.Device",
                        on_delete=django.db.models.deletion.CASCADE,
                    ),
                ),
            ],
            bases=("nsot.resource",),
        ),
        migrations.AddField(
            model_name="assignment",
            name="interface",
            field=models.ForeignKey(to="nsot.Interface", on_delete=django.db.models.deletion.CASCADE),
        ),
        migrations.AlterUniqueTogether(
            name="interface",
            unique_together=set([("device", "name")]),
        ),
        migrations.AlterIndexTogether(
            name="interface",
            index_together=set([("device", "name")]),
        ),
        migrations.AlterUniqueTogether(
            name="assignment",
            unique_together=set([("address", "interface")]),
        ),
    ]



import copy
from rest_framework.routers import DefaultRouter, SimpleRouter

__all__ = [
    "BulkRouter",
]


class BulkRouter(DefaultRouter):
    """
    Map http methods to actions defined on the bulk mixins.
    """

    routes = copy.deepcopy(SimpleRouter.routes)
    routes[0].mapping.update(
        {
            "put": "bulk_update",
            "patch": "partial_bulk_update",
            "delete": "bulk_destroy",
        }
    )



# -*- coding: utf-8 -*-

from django.db import models, migrations
import django.db.models.deletion

class Migration(migrations.Migration):

    dependencies = [
        ("nsot", "0013_auto_20151002_1443"),
    ]

    operations = [
        migrations.AlterField(
            model_name="interface",
            name="parent",
            field=models.ForeignKey(
                related_name="children",
                on_delete=django.db.models.deletion.PROTECT,
                default=None,
                to="nsot.Interface",
                blank=True,
                help_text="Unique ID of the parent Interface.",
                null=True,
                verbose_name="Parent",
            ),
        ),
    ]



"""
Django settings for nsot project.
"""

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)

import os
import re
import sys

import macaddress
from netaddr import eui

from nsot.version import __version__

# So we can emit this in the API.
NSOT_VERSION = __version__

# Path where the code is found. (aka project root)
BASE_DIR = os.path.normpath(os.path.join(os.path.dirname(__file__), os.pardir))

# Path where the config file is found.
CONF_ROOT = os.path.abspath(os.path.dirname(__file__))

# A boolean that turns on/off debug mode. Never deploy a site into production
# with DEBUG turned on.
# Default: False
DEBUG = False

DEFAULT_AUTO_FIELD = "django.db.models.AutoField"
TIME_ZONE = "UTC"
USE_TZ = True

#################
# Core Settings #
#################

# A tuple of strings designating all applications that are enabled in this
# Django installation.
INSTALLED_APPS = (
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django_extensions",
    "django_filters",
    "guardian",
    "rest_framework",
    "nsot",
)

# The model to use to represent a User.
AUTH_USER_MODEL = "nsot.User"
ANONYMOUS_USER_NAME = "anonymous@service.local"

# A tuple of authentication backend classes (as strings) to use when attempting
# to authenticate a user.
# https://docs.djangoproject.com/en/1.8/topics/auth/customizing/#authentication-backends
AUTHENTICATION_BACKENDS = (
    "nsot.middleware.auth.EmailHeaderBackend",
    "django.contrib.auth.backends.ModelBackend",
    "nsot.middleware.auth.NsotObjectPermissionsBackend",
)

# A tuple of middleware classes to use.
# https://docs.djangoproject.com/en/1.8/topics/http/middleware/
MIDDLEWARE = (
    "nsot.middleware.request_logging.LoggingMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "nsot.middleware.auth.EmailHeaderMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "django.middleware.security.SecurityMiddleware",
)

# A string representing the full Python import path to your root URLconf.
ROOT_URLCONF = "nsot.conf.urls"

# The full Python path of the WSGI application object that Django's built-in
# servers (e.g. runserver) will use.
WSGI_APPLICATION = "nsot.wsgi.application"

# When set to True, if the request URL does not match any of the patterns in
# the URLconf and it doesn't end in a slash, an HTTP redirect is issued to the
# same URL with a slash appended. Note that the redirect may cause any data
# submitted in a POST request to be lost.
# Default: True
APPEND_SLASH = True

# Template loaders for Django admin, DRF browsable API, etc.
TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [
            os.path.join(BASE_DIR, "templates").replace("\\", "/"),
        ],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.csrf",
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

##################
# REST Framework #
##################
# Release version of the API to publish.
NSOT_API_VERSION = "1.0"

# Settings for Django REST Framework (DRF)
REST_FRAMEWORK = {
    "DEFAULT_FILTER_BACKENDS": (
        "django_filters.rest_framework.DjangoFilterBackend",
    ),
    "DEFAULT_RENDERER_CLASSES": [
        "rest_framework.renderers.JSONRenderer",
        "nsot.api.renderers.FilterlessBrowsableAPIRenderer",
    ],
    "DEFAULT_PAGINATION_CLASS": "rest_framework.pagination.LimitOffsetPagination",
    "DEFAULT_VERSIONING_CLASS": "rest_framework.versioning.AcceptHeaderVersioning",
    "DEFAULT_VERSION": NSOT_API_VERSION,
    "DEFAULT_PERMISSION_CLASSES": (
        "rest_framework.permissions.DjangoModelPermissions",
    ),
    "DEFAULT_AUTHENTICATION_CLASSES": (
        "nsot.api.auth.AuthTokenAuthentication",
        "nsot.api.auth.EmailHeaderAuthentication",
        "rest_framework.authentication.BasicAuthentication",
        "rest_framework.authentication.SessionAuthentication",
    ),
    "EXCEPTION_HANDLER": "nsot.exc.custom_exception_handler",
    "TEST_REQUEST_DEFAULT_FORMAT": "json",
    "PAGE_SIZE": None,  # No default pagination.
}

############
# Database #
############
# https://docs.djangoproject.com/en/1.8/ref/settings/#databases

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": os.path.join(BASE_DIR, "nsot.sqlite3"),
    }
}

#########
# Cache #
#########
# NSoT includes supportfor caching of API results, however the default is to
# use to the "dummy" cache that doesn't actually cache -- it just implements
# the cache interface without doing anything.
#
# If you need caching, see the docs to choose a caching backend:
# https://docs.djangoproject.com/en/1.8/ref/settings/#caches
CACHES = {
    "default": {
        "BACKEND": "django.core.cache.backends.dummy.DummyCache",
    }
}

###############
# Application #
###############

#
# Back-end
#

# The type of workers to use.
# http://docs.gunicorn.org/en/latest/settings.html#worker-class
# Default: 'gevent'
NSOT_WORKER_CLASS = "gthread"

# Load application code before the worker processes are forked.
# http://docs.gunicorn.org/en/latest/settings.html#preload-app
# Default: False
NSOT_PRELOAD = False

# The maximum number of requests a worker will process before restarting.
# http://docs.gunicorn.org/en/latest/settings.html#max-requests
# Default: 0
NSOT_MAX_REQUESTS = 0

# The maximum jitter to add to the max_requests setting. The jitter causes the
# restart per worker to be randomized by randint(0, max_requests_jitter). This
# is intended to stagger worker restarts to avoid all workers restarting at the
# same time.
# http://docs.gunicorn.org/en/latest/settings.html#max-requests-jitter
# Default: 0
NSOT_MAX_REQUESTS_JITTER = 0

#
# User-configurable
#

# The address on which the application will listen.
# Default: 'localhost'
NSOT_HOST = "localhost"

# The port on which the application will be accessed.
# Default: 8990
NSOT_PORT = 8990

# The number of gunicorn worker processes for handling requests.
# Default: 4
NSOT_NUM_WORKERS = 4

# Timeout in seconds before gunicorn workers are killed/restarted.
# Default: 30
NSOT_WORKER_TIMEOUT = 30

# If True, serve static files directly from the app.
# Default: True
SERVE_STATIC_FILES = True

############
# Security #
############

# Header to check for Authenticated Email. This is intended for use behind an
# authenticating reverse proxy.
USER_AUTH_HEADER = "X-NSoT-Email"

# The age, in seconds, until an AuthToken granted by the API will expire.
# Default: 600
AUTH_TOKEN_EXPIRY = 600  # 10 minutes

# A list of strings representing the host/domain names that this Django site
# can serve. This is a security measure to prevent an attacker from poisoning
# caches and triggering password reset emails with links to malicious hosts by
# submitting requests with a fake HTTP Host header, which is possible even
# under many seemingly-safe web server configurations.
# https://docs.djangoproject.com/en/1.8/ref/settings/#allowed-hosts
ALLOWED_HOSTS = ["*"]

# The name of the cookie to use for the CSRF authentication token.
# Default: 'csrftoken'
CSRF_COOKIE_NAME = "_xsrf"

# Coming in Django 1.9
# The name of the request header used for CSRF authentication.
# https://github.com/django/django/commit/668d53c
# Default: 'X-CSRFToken'
# CSRF_HEADER_NAME = 'X-XSRFToken'

# Whether to create new users as superusers. If True, when new users are created
# they will be given superuser permissions. If False, new users will not be
# given superuser and it will be expected that permissions will be defined to
# control admin privileges. This only applies to users automatically created
# when authenticated using the "auth header" method, which is the default.
# Default: True
NSOT_NEW_USERS_AS_SUPERUSER = True

############
# Guardian #
############

# This setting is to silence any warnings raised by Guardian due to the usage
# of a custom auth backend implementation as opposed to
# guardian.backends.ObjectPermissionBackend
SILENCED_SYSTEM_CHECKS = ["guardian.W001"]

# The Guardian anonymous user is different to the Django Anonymous user.
# The Django Anonymous user does not have an entry in the database,
# however the Guardian anonymous user does.
ANONYMOUS_USER_NAME = "anonymous@service.local"

################
# Static files #
################

# (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.8/howto/static-files/

# URL to use when referring to static files located in STATIC_ROOT.
# Default: '/static/'
STATIC_URL = "/static/"

# This setting defines the additional locations the staticfiles app will
# traverse if the FileSystemFinder finder is enabled, e.g. if you use the
# collectstatic or findstatic management command or use the static file serving
# view.
# Default: $BASE_DIR/static
STATICFILES_DIRS = (os.path.join(BASE_DIR, "static"),)

# The absolute path to the directory where collectstatic will collect static
# files for deployment.
# Default: $BASE_DIR/staticfiles'
STATIC_ROOT = os.path.realpath(os.path.join(BASE_DIR, "staticfiles"))

###########
# Swagger #
###########

SWAGGER_SETTINGS = {
    "exclude_namespaces": ["index"],
}

###########
# Logging #
###########
LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "verbose": {
            "format": "[%(asctime)s] [%(process)d] [%(levelname)s] %(message)s",
            "datefmt": "%Y-%m-%d %H:%M:%S %z",
        },
        "simple": {"format": "%(levelname)s %(message)s"},
    },
    "handlers": {
        "console": {
            "level": "DEBUG",
            "class": "logging.StreamHandler",
            "stream": sys.stdout,
            "formatter": "verbose",
        },
    },
    "loggers": {
        "django.request": {
            "handlers": ["console"],
            "propagate": True,
            "level": "ERROR",
        },
        "nsot": {
            "handlers": ["console"],
            "level": "INFO",
        },
        "rest_framework": {
            "handlers": ["console"],
            "level": "ERROR",
        },
        "nsot_server": {
            "handlers": ["console"],
            "level": "INFO",
        },
    },
}

##############
# Attributes #
##############

# Acceptable regex pattern for naming Attribute objects.
ATTRIBUTE_NAME = re.compile(r"^[a-z][a-z0-9_]*$")

###########
# Devices #
###########

# Acceptable regex pattern for naming Device objects.
# Single DNS label (RFC 952/1123): 1-63 chars, alphanumeric ends, hyphens ok
_LABEL = r"[A-Za-z0-9](?:[A-Za-z0-9\-]{0,61}[A-Za-z0-9])?"
# Device name: one or more dot-separated labels (FQDN without trailing dot).
# DEVICE_NAME_PATTERN is the raw pattern (no anchors) reused by DeviceViewSet's
# lookup_value_regex for URL routing.
DEVICE_NAME_PATTERN = rf"{_LABEL}(?:\.{_LABEL})*"
DEVICE_NAME = re.compile(rf"^{DEVICE_NAME_PATTERN}$")

##############
# Interfaces #
##############

# The default format for displaying MAC addresses. This defaults to
# ":"-separated and expanded (e.g. '00:00:00:00:00:00')
MACADDRESS_DEFAULT_DIALECT = "macaddress.mac_linux"

# The default speed in Mbps for newly device interfaces if not otherwise
# specified.
INTERFACE_DEFAULT_SPEED = 1000  # In Mbps (e.g. 1Gbps)

# Default MAC address ('00:00:00:00:00:00')
# TODO(jathan): Dialect is hard-coded instead of reading from
# ``macaddress.default_dialect()`` because of a race condition in evaluating
# settings in Django >=2.x
INTERFACE_DEFAULT_MAC = eui.EUI(0, dialect=macaddress.mac_linux, version=48)

# These are mappings to the formal integer types from SNMP IF-MIB::ifType. The
# types listed here are the most commonly found in the wild.
#
# *IF YOU ARE GOING TO MODIFY THIS*: This MUST be a list of 2-tuples of
# (iftype, name), where iftype is the unique ID for the IANA ifType SNMP MIB,
# and name is whatever name your little heart desires, but hopefully matches
# the legit description according to the MIB.
#
# Ref: https://www.iana.org/assignments/ianaiftype-mib/ianaiftype-mib
INTERFACE_TYPE_CHOICES = (
    (6, "ethernet"),  # for all ethernet-like interfaces
    (1, "other"),  # none of the following
    (135, "l2vlan"),  # Layer 2 Virtual LAN using 802.1Q
    (136, "l3vlan"),  # Layer 3 Virtual LAN using IP
    (161, "lag"),  # IEEE 802.3ad Link Aggregat
    (24, "loopback"),  # softwareLoopback
    (150, "mpls"),  # MPLS Tunnel Virtual Interface
    (53, "prop_virtual"),  # proprietary virtual/internal
    (131, "tunnel"),  # Encapsulation interface
)

INTERFACE_DEFAULT_TYPE = 6  # ethernet

############
# Networks #
############

# How long is an interconnect? (aka a point-to-point link)
NETWORK_INTERCONNECT_PREFIXES = (31, 127)

# CIDR prefix lengths for host addresses
HOST_PREFIXES = (32, 128)

# Valid IP versions
IP_VERSIONS = ("4", "6")

# Whether to compress IPv6 for display purposes, for example:
# - Exploded (default): 2620:0100:6000:0000:0000:0000:0000:0000/40
# - Compressed: 2620:100:6000::/40
# Default: True
NSOT_COMPRESS_IPV6 = True



"""
Constants for working with models.
"""

from collections import OrderedDict

from django.conf import settings

# These are constants that becuase they are tied directly to the underlying
# objects are explicitly NOT USER CONFIGURABLE.
RESOURCE_BY_IDX = (
    "Site",
    "Network",
    "Attribute",
    "Device",
    "Interface",
    "Circuit",
    "Protocol",
    "ProtocolType",
)
RESOURCE_BY_NAME = OrderedDict(
    (obj_type, idx) for idx, obj_type in enumerate(RESOURCE_BY_IDX)
)

# Valid change event types
CHANGE_EVENTS = ("Create", "Update", "Delete")

# Resource objects that can be used for Change events
VALID_CHANGE_RESOURCES = tuple(RESOURCE_BY_IDX)

# Resource objects that can have attributes
VALID_ATTRIBUTE_RESOURCES = (
    "Network",
    "Device",
    "Interface",
    "Circuit",
    "Protocol",
)

# Lists of 2-tuples of (value, option) for displaying choices in certain model
# serializer/form fields.
CHANGE_RESOURCE_CHOICES = [(c, c) for c in VALID_CHANGE_RESOURCES]
EVENT_CHOICES = [(c, c) for c in CHANGE_EVENTS]
IP_VERSION_CHOICES = [(c, c) for c in settings.IP_VERSIONS]
RESOURCE_CHOICES = [(c, c) for c in VALID_ATTRIBUTE_RESOURCES]

# Unique interface type IDs.
INTERFACE_TYPES = [t[0] for t in settings.INTERFACE_TYPE_CHOICES]

# Reverse mapping from type name to type ID (e.g. "ethernet" -> 6).
INTERFACE_TYPE_BY_NAME = {
    name: id for id, name in settings.INTERFACE_TYPE_CHOICES
}



import ipaddress
import logging

from django.conf import settings
from django.db import models
from django.utils.datastructures import DictWrapper
from django_extensions.db.fields.json import JSONField
from macaddress.fields import MACAddressField as BaseMACAddressField

from . import exc

__all__ = ("BinaryIPAddressField", "JSONField", "MACAddressField")

log = logging.getLogger(__name__)


class BinaryIPAddressField(models.Field):
    """IP Address field that stores values as varbinary."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.editable = True

    def db_type(self, connection):
        engine = connection.settings_dict["ENGINE"]

        # Use the native 'inet' type for Postgres.
        if "postgres" in engine:
            return "inet"

        # Or 'varbinary' for everyone else.
        data = DictWrapper(self.__dict__, connection.ops.quote_name, "qn_")
        return "varbinary(%(max_length)s)" % data

    def _parse_ip_address(self, value):
        try:
            obj = ipaddress.ip_address(str(value))
        except ValueError:
            obj = ipaddress.ip_address(bytes(value))

        # Display IPv6 as compressed or not? This is a no-op vor IPv4.
        if settings.NSOT_COMPRESS_IPV6:
            return obj.compressed

        return obj.exploded

    def from_db_value(self, value, expression, connection):
        """DB -> Python."""
        if value is None:
            return value

        return self._parse_ip_address(value)

    def to_python(self, value):
        """Object -> Python."""
        if isinstance(value, (ipaddress.IPv4Address, ipaddress.IPv6Address)):
            return value

        if value is None:
            return value

        return self._parse_ip_address(value)

    def get_db_prep_value(self, value, connection, prepared=False):
        """Python -> DB."""
        # To account for null defaults when performing migrations
        if value is None:
            return None

        engine = connection.settings_dict["ENGINE"]

        # Send the value as-is to Postgres.
        if "postgres" in engine:
            return value

        # Or packed binary for everyone else.
        return ipaddress.ip_address(value).packed


class MACAddressField(BaseMACAddressField):
    """
    Subclass of base field to raise a DRF ValidationError.

    DRF handles Django's default ValidationError, but this is so that we can
    always expect the DRF version, for better consistency in debugging and
    testing.
    """

    def from_db_value(self, value, expression, connection):
        # If value is an integer that is a string, make it an int
        if isinstance(value, str) and value.isdigit():
            value = int(value)

        try:
            return super().from_db_value(value, expression, connection)
        except exc.DjangoValidationError as err:
            raise exc.ValidationError(str(err))

    def to_python(self, value):
        # If value is an integer that is a string, make it an int
        if isinstance(value, str) and value.isdigit():
            value = int(value)

        try:
            return super().to_python(value)
        except exc.DjangoValidationError as err:
            raise exc.ValidationError(str(err))



# -*- coding: utf-8 -*-

from django.db import models, migrations
import nsot.fields

class Migration(migrations.Migration):

    dependencies = [
        ("nsot", "0009_auto_20150811_1245"),
    ]

    operations = [
        migrations.AlterField(
            model_name="attribute",
            name="description",
            field=models.CharField(default="", max_length=255, blank=True),
        ),
        migrations.AlterField(
            model_name="interface",
            name="description",
            field=models.CharField(default="", max_length=255, blank=True),
        ),
        migrations.AlterField(
            model_name="interface",
            name="mac_address",
            field=nsot.fields.MACAddressField(
                default=0,
                blank=True,
                help_text="If not provided, defaults to 00:00:00:00:00:00.",
                integer=True,
                null=True,
                verbose_name="MAC Address",
                db_index=True,
            ),
        ),
        migrations.AlterField(
            model_name="interface",
            name="name",
            field=models.CharField(
                help_text="The name of the interface as it appears on the device.",
                max_length=255,
                db_index=True,
            ),
        ),
        migrations.AlterField(
            model_name="interface",
            name="speed",
            field=models.IntegerField(
                default=10000,
                help_text="Integer of Mbps of interface (e.g. 20000 for 20 Gbps). If not provided, defaults to 10000.",
                db_index=True,
                blank=True,
            ),
        ),
        migrations.AlterField(
            model_name="interface",
            name="type",
            field=models.IntegerField(
                default=6,
                help_text="If not provided, defaults to 'ethernet'.",
                db_index=True,
                verbose_name="Interface Type",
                choices=[
                    (6, b"ethernet"),
                    (1, b"other"),
                    (135, b"l2vlan"),
                    (136, b"l3vlan"),
                    (161, b"lag"),
                    (24, b"loopback"),
                    (150, b"mpls"),
                    (131, b"tunnel"),
                ],
            ),
        ),
    ]



# -*- coding: utf-8 -*-

from django.db import models, migrations

class Migration(migrations.Migration):

    dependencies = [
        ("nsot", "0008_auto_20150811_1222"),
    ]

    operations = [
        migrations.AlterField(
            model_name="interface",
            name="type",
            field=models.IntegerField(
                default=6,
                db_index=True,
                verbose_name="Interface Type",
                choices=[
                    (1, "other"),
                    (6, "ethernet"),
                    (24, "loopback"),
                    (131, "tunnel"),
                    (135, "l2vlan"),
                    (136, "l3vlan"),
                    (150, "mpls"),
                    (161, "lag"),
                ],
            ),
        ),
    ]



# -*- coding: utf-8 -*-

from django.db import models, migrations
import django.db.models.deletion

class Migration(migrations.Migration):

    dependencies = [
        ("nsot", "0007_auto_20150811_1201"),
    ]

    operations = [
        migrations.AlterField(
            model_name="assignment",
            name="address",
            field=models.ForeignKey(
                related_name="assignments",
                to="nsot.Network",
                on_delete=django.db.models.deletion.CASCADE,
            ),
        ),
    ]



# -*- coding: utf-8 -*-

from django.db import models, migrations
import django.db.models.deletion

class Migration(migrations.Migration):

    dependencies = [
        ("nsot", "0011_auto_20150930_1557"),
    ]

    operations = [
        migrations.AlterField(
            model_name="interface",
            name="description",
            field=models.CharField(
                default="",
                help_text="A brief yet helpful description.",
                max_length=255,
                blank=True,
            ),
        ),
        migrations.AlterField(
            model_name="interface",
            name="device",
            field=models.ForeignKey(
                related_name="interfaces",
                verbose_name="Device",
                to="nsot.Device",
                on_delete=django.db.models.deletion.CASCADE,
                help_text="Unique ID of the connected Device.",
            ),
        ),
        migrations.AlterField(
            model_name="interface",
            name="parent",
            field=models.ForeignKey(
                related_name="children",
                on_delete=django.db.models.deletion.PROTECT,
                default=None,
                to="nsot.Interface",
                blank=True,
                help_text="Unique ID of the parent Interface.",
                null=True,
                verbose_name="Parent",
            ),
        ),
        migrations.AlterField(
            model_name="interface",
            name="type",
            field=models.IntegerField(
                default=6,
                help_text="If not provided, defaults to 'ethernet'.",
                db_index=True,
                verbose_name="Interface Type",
                choices=[
                    (6, b"ethernet"),
                    (1, b"other"),
                    (135, b"l2vlan"),
                    (136, b"l3vlan"),
                    (161, b"lag"),
                    (24, b"loopback"),
                    (150, b"mpls"),
                    (53, b"prop_virtual"),
                    (131, b"tunnel"),
                ],
            ),
        ),
    ]



# -*- coding: utf-8 -*-

from django.db import models, migrations
import django_extensions.db.fields.json

def migrate_network_fields(apps, schema_editor):
    """
    Migrate new Network fields.
    """
    Network = apps.get_model("nsot", "Network")
    Network_temp = apps.get_model("nsot", "Network_temp")
    for net in Network.objects.iterator():
        net_tmp = Network_temp.objects.create(
            network_address=net.network_address,
            broadcast_address=net.broadcast_address,
            prefix_length=net.prefix_length,
            ip_version=net.ip_version,
            is_ip=net.is_ip,
            id=net.resource_ptr_id,
            parent_id=net.parent_id,
            _attributes_cache=net._attributes,
            site=net.site,
        )

class Migration(migrations.Migration):

    dependencies = [
        ("nsot", "0016_move_device_data"),
    ]

    operations = [
        # Network _attributes_cache, new_id
        migrations.RunPython(migrate_network_fields),
    ]



from django.db import models

from .. import exc


class ProtocolType(models.Model):
    """
    Representation of protocol types (e.g. bgp, is-is, ospf, etc.)
    """

    name = models.CharField(
        max_length=16,
        db_index=True,
        help_text="Name of this type of protocol (e.g. OSPF, BGP, etc.)",
    )
    description = models.CharField(
        max_length=255,
        default="",
        blank=True,
        null=False,
        help_text="A description for this ProtocolType",
    )
    required_attributes = models.ManyToManyField(
        "Attribute",
        db_index=True,
        related_name="protocol_types",
        help_text=(
            "All Attributes which are required by this ProtocolType. If a"
            " Protocol of this type is saved and is missing one of these"
            " attributes, a ValidationError will be raised."
        ),
    )
    site = models.ForeignKey(
        "Site",
        db_index=True,
        related_name="protocol_types",
        on_delete=models.PROTECT,
        verbose_name="Site",
        help_text="Unique ID of the Site this ProtocolType is under.",
    )

    def __str__(self):
        return "%s" % self.name

    class Meta:
        unique_together = ("site", "name")

    def get_required_attributes(self):
        """Return a list of the names of ``self.required_attributes``."""
        # FIXME(jathan): These should probably cached on the model and updated
        # on write. Revisit after we see how performance plays out in practice.
        return list(self.required_attributes.values_list("name", flat=True))

    def to_dict(self):
        return {
            "id": self.id,
            "name": self.name,
            "description": self.description,
            "required_attributes": self.get_required_attributes(),
            "site": self.site_id,
        }


# Signals
def required_attributes_changed(
    sender, instance, action, reverse, model, pk_set, **kwargs
):
    """
    Signal handler that disallows anything but Protocol attributes to be added
    to a ProtocolType.required_attributes.
    """
    if action == "pre_add":
        # First filter in Protocol attributes.
        attrs = model.objects.filter(pk__in=pk_set)
        if attrs.exclude(resource_name="Protocol").exists():
            raise exc.ValidationError(
                {"required_attributes": "Only Protocol attributes are allowed"}
            )

        # Then make sure that they match the site of the incoming instance.
        wrong_site = attrs.exclude(site_id=instance.site_id)
        if wrong_site.exists():
            bad_attrs = [str(w) for w in wrong_site]
            raise exc.ValidationError(
                {
                    "required_attributes": (
                        "Attributes must share the same site as "
                        "ProtocolType.site. Got: %s" % bad_attrs
                    )
                }
            )


# Register required_attributes_changed -> ProtocolType.required_attributes
models.signals.m2m_changed.connect(
    required_attributes_changed,
    sender=ProtocolType.required_attributes.through,
)



from django.db import models

from .. import exc, util
from .resource import Resource


class Circuit(Resource):
    """Represents two network Interfaces that are connected"""

    # Derive site from A-side endpoint
    site_source_field = "endpoint_a"

    # A-side endpoint interface
    endpoint_a = models.OneToOneField(
        "Interface",
        on_delete=models.PROTECT,
        db_index=True,
        null=False,
        related_name="circuit_a",
        verbose_name="A-side endpoint Interface",
        help_text="Unique ID of Interface at the A-side.",
    )

    # Z-side endpoint interface
    endpoint_z = models.OneToOneField(
        "Interface",
        on_delete=models.PROTECT,
        db_index=True,
        null=True,
        related_name="circuit_z",
        verbose_name="Z-side endpoint Interface",
        help_text="Unique ID of Interface at the Z-side.",
    )

    # We are currently inferring the site_id from the parent (A-side) Interface
    # in the .save() method
    site = models.ForeignKey(
        "Site",
        db_index=True,
        related_name="circuits",
        on_delete=models.PROTECT,
        help_text="Unique ID of the Site this Circuit is under.",
    )

    name = models.CharField(
        max_length=255,
        unique=True,
        default="",
        help_text=(
            "Unique display name of the Circuit. If not provided, defaults to "
            "'{device_a}:{interface_a}_{device_z}:{interface_z}'"
        ),
    )

    # This doesn't use the built-in SlugField since we're doing our own
    # slugification (django.utils.text.slugify() is too agressive)
    name_slug = models.CharField(
        db_index=True,
        editable=False,
        max_length=255,
        null=True,
        unique=True,
        help_text=(
            "Slugified version of the name field, used for the natural key"
        ),
    )

    def __str__(self):
        return "%s" % self.name

    class Meta:
        # TODO(jathan): Benchmark queries on a large database to identify
        # whether we need explicit indices for this model. In my initial
        # testing all of the common lookup fields are already indexed so this
        # may not be necessary.
        """
        index_together = [
            ('endpoint_a', 'endpoint_z'),
            ('site', 'name_slug'),
        ]
        """

    @property
    def interfaces(self):
        """Return interfaces associated with this circuit."""
        intf_list = [self.endpoint_a, self.endpoint_z]
        return [i for i in intf_list if i is not None]

    @property
    def addresses(self):
        """Return addresses associated with this circuit.

        This includes addresses associated with all descendant interfaces.
        """
        addresses = []
        for interface in self.interfaces:
            addresses.extend(interface.addresses.all())

            # For each interface, get addresses of all descendant interfaces
            # (children, grandchildren, etc.) and extend the list.
            for descendant in interface.get_descendants():
                addresses.extend(descendant.addresses.all())

        return addresses

    @property
    def devices(self):
        """Return devices associated with this circuit."""
        return [i.device for i in self.interfaces]

    def interface_for(self, device):
        """
        Given a Device object, return the interface attached to this Circuit
        which belongs to that Device. If both ends of the Circuit are attached
        to the Device, the A-side is returned.

        If neither ends of this Circuit are attached to Device, then None is
        returned
        """
        if self.endpoint_a.device == device:
            return self.endpoint_a
        if self.endpoint_z and self.endpoint_z.device == device:
            return self.endpoint_z
        return None

    def clean_endpoint_a(self, value):
        if Circuit.objects.filter(endpoint_z=value).exists():
            raise exc.ValidationError(
                {"endpoint_a": "Interface already used as an endpoint_z"}
            )

        return self.endpoint_a

    def clean_endpoint_z(self, value):
        if Circuit.objects.filter(endpoint_a=value).exists():
            raise exc.ValidationError(
                {"endpoint_z": "Interface already used as an endpoint_a"}
            )

        return self.endpoint_z

    def clean_name(self, value):
        if value:
            return value

        # Add display name of hostname:intf_hostname:intf
        name = f"{self.endpoint_a}_{self.endpoint_z}"
        return name

    def clean_fields(self, exclude=None):
        self.site_id = self.clean_site(self.site_id)
        self.endpoint_a = self.clean_endpoint_a(self.endpoint_a_id)
        self.endpoint_z = self.clean_endpoint_z(self.endpoint_z_id)
        self.name = self.clean_name(self.name)

        self.name_slug = util.slugify(self.name)

    def to_dict(self):
        return {
            "id": self.id,
            "site_id": self.site_id,
            "name": self.name,
            "name_slug": self.name_slug,
            "endpoint_a": self.endpoint_a and self.endpoint_a.name_slug,
            "endpoint_z": self.endpoint_z and self.endpoint_z.name_slug,
            "attributes": self.get_attributes(),
        }



"""
Gettings stats out of NSoT.
"""

from netaddr import IPNetwork, IPSet

__all__ = ("calculate_network_utilization", "get_network_utilization")


def calculate_network_utilization(parent, hosts, as_string=False):
    """
    Calculate utilization for a network and its descendants.

    :param parent:
        The parent network

    :param hosts:
        List of host IPs descendant from parent

    :param as_string:
        Whether to return stats as a string
    """
    parent = IPNetwork(str(parent))
    hosts = IPSet(str(ip) for ip in hosts if IPNetwork(str(ip)) in parent)

    used = float(hosts.size) / float(parent.size)
    free = 1 - used
    num_free = parent.size - hosts.size

    stats = {
        "percent_used": used,
        "num_used": hosts.size,
        "percent_free": free,
        "num_free": num_free,
        "max": parent.size,
    }

    # 10.47.216.0/22 - 14% used (139), 86% free (885)
    if as_string:
        return f"{parent} - {used:.0%} used ({hosts.size}), {free:.0%} free ({num_free})"

    return stats


def get_network_utilization(network, as_string=False):
    """
    Get utilization from Network instance.

    :param network:
        A Network model instance

    :param as_string:
        Whether to return stats as a string
    """
    descendants = network.get_descendants().filter(is_ip=True)
    return calculate_network_utilization(network, descendants, as_string)



# -*- coding: utf-8 -*-

from django.db import migrations, models
import django.db.models.deletion
import django_extensions.db.fields.json

class Migration(migrations.Migration):

    dependencies = [
        ("nsot", "0035_fix_interface_name_slug"),
    ]

    operations = [
        migrations.CreateModel(
            name="Protocol",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                (
                    "_attributes_cache",
                    django_extensions.db.fields.json.JSONField(
                        help_text="Local cache of attributes. (Internal use only)",
                        blank=True,
                    ),
                ),
                (
                    "auth_string",
                    models.CharField(
                        default="",
                        help_text="Authentication string (such as MD5 sum)",
                        max_length=255,
                        blank=True,
                    ),
                ),
                (
                    "description",
                    models.CharField(
                        default="",
                        help_text="Description for this Protocol",
                        max_length=255,
                        blank=True,
                    ),
                ),
                (
                    "circuit",
                    models.ForeignKey(
                        related_name="protocols",
                        blank=True,
                        to="nsot.Circuit",
                        on_delete=django.db.models.deletion.CASCADE,
                        help_text="Circuit that this protocol is running over.",
                        null=True,
                    ),
                ),
                (
                    "device",
                    models.ForeignKey(
                        related_name="protocols",
                        to="nsot.Device",
                        on_delete=django.db.models.deletion.CASCADE,
                        help_text="Device that this protocol is running on",
                    ),
                ),
                (
                    "interface",
                    models.ForeignKey(
                        related_name="protocols",
                        blank=True,
                        to="nsot.Interface",
                        on_delete=django.db.models.deletion.CASCADE,
                        help_text="Interface this protocol is running on. Either interface or circuit must be populated.",
                        null=True,
                    ),
                ),
                (
                    "site",
                    models.ForeignKey(
                        related_name="protocols",
                        on_delete=django.db.models.deletion.PROTECT,
                        blank=True,
                        to="nsot.Site",
                        help_text="Unique ID of the Site this Protocol is under. If not set, this be inherited off of the device's site",
                        null=True,
                        verbose_name="Site",
                    ),
                ),
            ],
            options={
                "ordering": ("device",),
            },
        ),
        migrations.CreateModel(
            name="ProtocolType",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="Name of this type of protocol (e.g. OSPF, BGP, etc.)",
                        unique=True,
                        max_length=16,
                        db_index=True,
                    ),
                ),
                (
                    "description",
                    models.CharField(
                        default="",
                        help_text="A description for this ProtocolType",
                        max_length=255,
                        blank=True,
                    ),
                ),
            ],
        ),
        migrations.AlterField(
            model_name="attribute",
            name="resource_name",
            field=models.CharField(
                help_text="The name of the Resource to which this Attribute is bound.",
                max_length=20,
                verbose_name="Resource Name",
                db_index=True,
                choices=[
                    ("Device", "Device"),
                    ("Interface", "Interface"),
                    ("Protocol", "Protocol"),
                    ("Network", "Network"),
                    ("Circuit", "Circuit"),
                ],
            ),
        ),
        migrations.AlterField(
            model_name="change",
            name="resource_name",
            field=models.CharField(
                help_text="The name of the Resource for this Change.",
                max_length=20,
                verbose_name="Resource Type",
                db_index=True,
                choices=[
                    ("Protocol", "Protocol"),
                    ("Network", "Network"),
                    ("Attribute", "Attribute"),
                    ("Site", "Site"),
                    ("Interface", "Interface"),
                    ("Circuit", "Circuit"),
                    ("Device", "Device"),
                ],
            ),
        ),
        migrations.AlterField(
            model_name="network",
            name="is_ip",
            field=models.BooleanField(
                default=False,
                help_text="Whether the Network is a host address or not.",
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="value",
            name="resource_name",
            field=models.CharField(
                help_text="The name of the Resource type to which the Value is bound.",
                max_length=20,
                verbose_name="Resource Type",
                db_index=True,
                choices=[
                    ("Protocol", "Protocol"),
                    ("Network", "Network"),
                    ("Attribute", "Attribute"),
                    ("Site", "Site"),
                    ("Interface", "Interface"),
                    ("Circuit", "Circuit"),
                    ("Device", "Device"),
                ],
            ),
        ),
        migrations.AddField(
            model_name="protocoltype",
            name="required_attributes",
            field=models.ManyToManyField(
                help_text="All Attributes which are required by this ProtocolType. If a Protocol of this type is saved and is missing one of these attributes, a ValidationError will be raised.",
                related_name="protocol_types",
                to="nsot.Attribute",
                db_index=True,
            ),
        ),
        migrations.AddField(
            model_name="protocol",
            name="type",
            field=models.ForeignKey(
                related_name="protocols",
                to="nsot.ProtocolType",
                on_delete=django.db.models.deletion.CASCADE,
                help_text="The type of this Protocol",
            ),
        ),
    ]



import logging
from collections import OrderedDict

from django.core.exceptions import (
    MultipleObjectsReturned,
    ObjectDoesNotExist,
)
from django.core.exceptions import (
    ValidationError as DjangoValidationError,
)
from django.db import IntegrityError
from django.db.models import ProtectedError
from rest_framework.exceptions import APIException, ValidationError
from rest_framework.views import exception_handler

log = logging.getLogger(__name__)

__all__ = (
    "BadRequest",
    "BaseHttpError",
    "Conflict",
    "DjangoValidationError",
    "Error",
    "Forbidden",
    "ModelError",
    "MultipleObjectsReturned",
    "NotFound",
    "ObjectDoesNotExist",
    "ProtectedError",
    "Unauthorized",
    "ValidationError",
)


def custom_exception_handler(exc, context):
    """Always handle errors all pretty-like."""
    # Call REST framework's default exception handler first to get the standard
    # error response.
    response = exception_handler(exc, context)

    # Now add the HTTP status code and message to the response.
    # We want an error response to look like this:
    # {
    #     "error": {
    #         "message": "Endpoint not found",
    #         "code": 404
    #     }
    # }
    log.debug("custom_exception_handler: exc = %r", exc)
    log.debug("custom_exception_handler: context = %r", context)
    if response is not None:
        orig_data = response.data
        log.debug("custom_exception_handler: orig_data = %r", orig_data)

        try:
            message = orig_data["detail"]
            if message == "Not found.":
                message = "Endpoint not found."
        except KeyError:
            message = orig_data
        except TypeError:
            message = orig_data[0]

        data = OrderedDict(
            [
                (
                    "error",
                    {
                        "message": message,
                        "code": response.status_code,
                    },
                ),
            ]
        )
        response.data = data

    request = context["request"]
    log.debug("custom_exception_handler: request = %r", request)
    log.debug("custom_exception_handler: dir(request) = %r", dir(request))
    log.debug("custom_exception_handler: request.data = %r", request.data)

    return response


class Error(APIException):
    """Baseclass for NSoT Exceptions."""


class ModelError(Error):
    """Base class for NSoT Model Exceptions."""


class BaseHttpError(Error):
    """Base HTTP error."""


class BadRequest(BaseHttpError):
    """HTTP 400 error."""

    status_code = 400


class Unauthorized(BaseHttpError):
    """HTTP 401 error."""

    status_code = 401


class Forbidden(BaseHttpError):
    """HTTP 403 error."""

    status_code = 403


class NotFound(BaseHttpError):
    """HTTP 404 error."""

    status_code = 404
    default_detail = "Endpoint not found."


class Conflict(BaseHttpError, IntegrityError):
    """HTTP 409 error."""

    status_code = 409



__version__ = "0.2.1"
__author__ = "Miroslav Shubernetskiy"

try:
    from .generics import *  # noqa
    from .mixins import *  # noqa
    from .serializers import *  # noqa
except Exception:
    pass



# -*- coding: utf-8 -*-

from django.db import models, migrations

def migrate_assignment_fields(apps, schema_editor):
    """Migrate new Assignment fields."""
    Assignment = apps.get_model("nsot", "Assignment")
    Assignment_temp = apps.get_model("nsot", "Assignment_temp")
    for asn in Assignment.objects.iterator():
        asn_tmp = Assignment_temp.objects.create(
            id=asn.id,
            address_id=asn.address_id,
            interface_id=asn.interface_id,
        )

class Migration(migrations.Migration):

    dependencies = [
        ("nsot", "0018_move_interface_data"),
    ]

    operations = [
        # Assignment id, address_id, interface_id
        migrations.RunPython(migrate_assignment_fields),
    ]



# -*- coding: utf-8 -*-

from django.db import migrations, models
import django.db.models.deletion

class Migration(migrations.Migration):

    dependencies = [
        ("nsot", "0036_add_protocol"),
    ]

    operations = [
        migrations.AddField(
            model_name="protocoltype",
            name="site",
            field=models.ForeignKey(
                related_name="protocol_types",
                on_delete=django.db.models.deletion.PROTECT,
                default=1,
                verbose_name="Site",
                to="nsot.Site",
                help_text="Unique ID of the Site this ProtocolType is under.",
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="change",
            name="resource_name",
            field=models.CharField(
                help_text="The name of the Resource for this Change.",
                max_length=20,
                verbose_name="Resource Type",
                db_index=True,
                choices=[
                    ("Protocol", "Protocol"),
                    ("Network", "Network"),
                    ("ProtocolType", "ProtocolType"),
                    ("Attribute", "Attribute"),
                    ("Site", "Site"),
                    ("Interface", "Interface"),
                    ("Circuit", "Circuit"),
                    ("Device", "Device"),
                ],
            ),
        ),
        migrations.AlterField(
            model_name="protocol",
            name="auth_string",
            field=models.CharField(
                default="",
                help_text="Authentication string (such as MD5 sum)",
                max_length=255,
                verbose_name="Auth String",
                blank=True,
            ),
        ),
        migrations.AlterField(
            model_name="protocoltype",
            name="name",
            field=models.CharField(
                help_text="Name of this type of protocol (e.g. OSPF, BGP, etc.)",
                max_length=16,
                db_index=True,
            ),
        ),
        migrations.AlterField(
            model_name="value",
            name="resource_name",
            field=models.CharField(
                help_text="The name of the Resource type to which the Value is bound.",
                max_length=20,
                verbose_name="Resource Type",
                db_index=True,
                choices=[
                    ("Device", "Device"),
                    ("Interface", "Interface"),
                    ("Protocol", "Protocol"),
                    ("Network", "Network"),
                    ("Circuit", "Circuit"),
                ],
            ),
        ),
        migrations.AlterUniqueTogether(
            name="protocoltype",
            unique_together=set([("site", "name")]),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2018-11-07 11:56

from django.db import migrations, models
import django_extensions.db.fields.json

class Migration(migrations.Migration):

    dependencies = [
        ("nsot", "0037_protocoltype_site__unique_together"),
    ]

    operations = [
        migrations.AlterField(
            model_name="interface",
            name="speed",
            field=models.IntegerField(
                blank=True,
                db_index=True,
                default=1000,
                help_text="Integer of Mbps of interface (e.g. 20000 for 20 Gbps). If not provided, defaults to 1000.",
                null=True,
            ),
        ),
    ]



from django.conf import settings
from django.urls import include, path

from . import routers, views

# Register all endpoints as a top-level resource
router = routers.BulkRouter(trailing_slash=settings.APPEND_SLASH)

# Resources pinned to API index at /
router.register(r"sites", views.SiteViewSet)
router.register(r"attributes", views.AttributeViewSet)
router.register(r"changes", views.ChangeViewSet)
router.register(r"circuits", views.CircuitViewSet)
router.register(r"devices", views.DeviceViewSet)
router.register(r"interfaces", views.InterfaceViewSet)
router.register(r"networks", views.NetworkViewSet)
router.register(r"protocols", views.ProtocolViewSet)
router.register(r"protocol_types", views.ProtocolTypeViewSet)
router.register(r"users", views.UserViewSet)
router.register(r"values", views.ValueViewSet)

# Nested router for resources under /sites
sites_router = routers.BulkNestedRouter(
    router, r"sites", lookup="site", trailing_slash=settings.APPEND_SLASH
)

# Resources that are nested under /sites
sites_router.register(r"attributes", views.AttributeViewSet)
sites_router.register(r"changes", views.ChangeViewSet)
sites_router.register(r"circuits", views.CircuitViewSet)
sites_router.register(r"devices", views.DeviceViewSet)
sites_router.register(r"interfaces", views.InterfaceViewSet)
sites_router.register(r"networks", views.NetworkViewSet)
sites_router.register(r"protocols", views.ProtocolViewSet)
sites_router.register(r"protocol_types", views.ProtocolTypeViewSet)
sites_router.register(r"values", views.ValueViewSet)

# Wire up our API using automatic URL routing.
# Additionally, we include login URLs for the browsable API.
urlpatterns = [
    # API routes
    path("", include(router.urls)),
    path("", include(sites_router.urls)),
    # Browsable API auth login
    path("auth/", include("rest_framework.urls", namespace="rest_framework")),
    # API auth_token login/verify (email/secret_key)
    path(
        "authenticate/",
        views.AuthTokenLoginView.as_view(),
        name="authenticate",
    ),
    path(
        "verify_token/",
        views.AuthTokenVerifyView.as_view(),
        name="verify_token",
    ),
]



"""
Validators for validating object fields.
"""

import ipaddress

import netaddr
from django.conf import settings
from django.core.validators import EmailValidator

from . import exc


def validate_mac_address(value):
    """Validate whether ``value`` is a valid MAC address."""
    if value is None:
        return value

    # If the incoming value is a string, cast it to an int
    if isinstance(value, str) and value.isdigit():
        value = int(value)

    # Directly invoke EUI object instead of using MACAddressField
    try:
        value = netaddr.EUI(value, version=48)
    except (ValueError, TypeError, netaddr.AddrFormatError):
        raise exc.ValidationError(
            {"mac_address": "Enter a valid MAC Address."}
        )

    return value


def validate_name(value):
    """Validate whether ``value`` is a valid name."""
    if not value:
        raise exc.ValidationError({"name": "This is a required field."})
    return value


def validate_cidr(value):
    """Validate whether ``value`` is a validr IPv4/IPv6 CIDR."""
    try:
        cidr = ipaddress.ip_network(str(value))
    except ValueError:
        raise exc.ValidationError(
            {
                "cidr": "%r does not appear to be an IPv4 or IPv6 network"
                % value
            }
        )
    else:
        return cidr


def validate_host_address(value):
    """Validate whether ``value`` is a host IP address."""
    cidr = validate_cidr(value)
    if cidr.prefixlen not in settings.HOST_PREFIXES:
        raise exc.ValidationError(
            {"address": "%r is not a valid host address!" % value}
        )
    return value


def validate_email(value):
    """Validate whether ``value`` is an email address."""
    validator = EmailValidator()
    try:
        validator(value)
    except exc.DjangoValidationError as err:
        raise exc.ValidationError({"email": str(err)})
    return value



"""
Command for running any pending upgrades.
"""

from django.core.management import call_command

from nsot.util.commands import NsotCommand


class Command(NsotCommand):
    help = "Performs any pending database migrations and upgrades"

    def add_arguments(self, parser):
        parser.add_argument(
            "--noinput",
            action="store_true",
            default=False,
            help="Tells Django to NOT prompt the user for input of any kind.",
        )

    def handle(self, **options):
        call_command(
            "migrate",
            interactive=(not options["noinput"]),
            traceback=options["traceback"],
            verbosity=options["verbosity"],
        )



from rest_framework_nested.routers import NestedSimpleRouter

from nsot.vendor.rest_framework_bulk.routes import BulkRouter

__all__ = ("BulkNestedRouter", "BulkRouter")

# Map of HTTP verbs to rest_framework_bulk operations.
BULK_OPERATIONS_MAP = {
    "put": "bulk_update",
    "patch": "partial_bulk_update",
    "delete": "bulk_destroy",
}


class BulkNestedRouter(NestedSimpleRouter):
    """
    Bulk-enabled nested router.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.routes[0].mapping.update(BULK_OPERATIONS_MAP)



# -*- coding: utf-8 -*-

from django.db import migrations, models

class Migration(migrations.Migration):

    dependencies = [
        ("nsot", "0031_populate_circuit_name_slug"),
    ]

    operations = [
        migrations.AlterField(
            model_name="change",
            name="change_at",
            field=models.DateTimeField(
                help_text="The timestamp of this Change.",
                auto_now_add=True,
                db_index=True,
            ),
        ),
        migrations.AlterIndexTogether(
            name="change",
            index_together=set(
                [("resource_name", "resource_id"), ("resource_name", "event")]
            ),
        ),
    ]



"""
WSGI config for nsot project.

It exposes the WSGI callable as a module-level variable named ``application``.

For more information on this file, see
https://docs.djangoproject.com/en/1.8/howto/deployment/wsgi/
"""

import os

from django.core.wsgi import get_wsgi_application

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "nsot.conf.settings")

from django.conf import settings

# If we're set to serve static files ourself (default), wrap the app w/ Cling
# (provided by dj-static).
if settings.SERVE_STATIC_FILES:
    from dj_static import Cling

    application = Cling(get_wsgi_application())
else:
    application = get_wsgi_application()



from django.db import models

from .. import exc, validators


class Assignment(models.Model):
    """
    DB object for assignment of addresses to interfaces (on devices).

    This is used to enforce constraints at the relationship level for addition
    of new address assignments.
    """

    address = models.ForeignKey(
        "Network",
        related_name="assignments",
        db_index=True,
        on_delete=models.CASCADE,
        help_text="Network to which this assignment is bound.",
    )
    interface = models.ForeignKey(
        "Interface",
        related_name="assignments",
        db_index=True,
        on_delete=models.CASCADE,
        help_text="Interface to which this assignment is bound.",
    )
    created = models.DateTimeField(auto_now_add=True)

    def __sr__(self):
        return "interface=%s, address=%s" % (self.interface, self.address)

    class Meta:
        unique_together = ("address", "interface")

    def clean_address(self, value):
        """Enforce that new addresses can only be host addresses."""
        addr = validators.validate_host_address(value)

        # Enforce uniqueness upon assignment.
        existing = Assignment.objects.filter(address=addr)
        if existing.filter(interface__device=self.interface.device).exists():
            raise exc.ValidationError(
                {"address": "Address already assigned to this Device."}
            )

        return value

    def clean_fields(self, exclude=None):
        self.clean_address(self.address)
        self.address.set_assigned()

    def save(self, *args, **kwargs):
        self.full_clean()
        super().save(*args, **kwargs)

    def to_dict(self):
        return {
            "id": self.id,
            "device": self.interface.device.id,
            "hostname": self.interface.device_hostname,
            "interface": self.interface.id,
            "interface_name": self.interface.name,
            "address": self.address.cidr,
        }


def revert_network_state_on_assignment_delete(sender, instance, **kwargs):
    """Revert Network state to ALLOCATED when no assignments remain."""
    from .network import Network

    try:
        address = Network.objects.get(pk=instance.address_id)
    except Network.DoesNotExist:
        return  # Network deleted too (e.g., _purge_addresses cascade)
    if not address.assignments.exists():
        address.set_allocated()


models.signals.post_delete.connect(
    revert_network_state_on_assignment_delete,
    sender=Assignment,
    dispatch_uid="revert_network_state_post_delete_assignment",
)



# -*- coding: utf-8 -*-

from django.db import models, migrations

class Migration(migrations.Migration):

    dependencies = [
        ("nsot", "0012_auto_20151002_1427"),
    ]

    operations = [
        migrations.AlterField(
            model_name="interface",
            name="name",
            field=models.CharField(
                help_text="The name of the interface as it appears on the Device.",
                max_length=255,
                db_index=True,
            ),
        ),
    ]



import ipaddress
import logging
import time
from operator import attrgetter

import netaddr
from django.conf import settings
from django.db import models

from .. import exc, fields, util, validators
from . import constants
from .resource import Resource, ResourceManager

log = logging.getLogger(__name__)


class NetworkManager(ResourceManager):
    """Manager for NetworkInterface objects."""

    def get_by_address(self, cidr, site=None):
        """
        Lookup a Network object by ``cidr``.

        :param cidr:
            IPv4/IPv6 CIDR string

        :param site:
            ``Site`` instance or ``site_id``
        """
        lookup_kwargs = util.cidr_to_dict(cidr)
        if site is not None:
            lookup_kwargs["site"] = site

        address = Network.objects.get(**lookup_kwargs)
        return address

    def get_closest_parent(self, cidr, prefix_length=0, site=None):
        """
        Return the closest matching parent Network for a ``cidr`` even if it
        doesn't exist in the database.

        :param cidr:
            IPv4/IPv6 CIDR string

        :param prefix_length:
            Maximum prefix length depth for closest parent lookup

        :param site:
            ``Site`` instance or ``site_id``
        """
        # Validate that it's a real CIDR
        cidr = validators.validate_cidr(cidr)
        broadcast_address = cidr.broadcast_address.exploded
        leaf = netaddr.IPNetwork(str(cidr))
        ip_version = leaf.version

        try:
            prefix_length = int(prefix_length)
        except ValueError:
            raise exc.ValidationError(
                {"prefix_length": "Invalid prefix_length: %r" % prefix_length}
            )

        # Walk the supernets backwrds from smallest to largest prefix.
        try:
            supernets = leaf.supernet(prefixlen=prefix_length)
        except ValueError as err:
            raise exc.ValidationError({"prefix_length": str(err)})
        else:
            supernets.reverse()

        # Enumerate all unique networks and prefixes
        network_addresses = {str(s.network) for s in supernets}
        prefix_lengths = {s.prefixlen for s in supernets}
        del supernets  # Free the memory because DevOps.

        # Prepare the queryset filter
        lookup_kwargs = {
            "network_address__in": network_addresses,
            "prefix_length__in": prefix_lengths,
            "ip_version": ip_version,
            "broadcast_address__gte": broadcast_address,
        }
        if site is not None:
            lookup_kwargs["site"] = site

        # Search for possible ancestors by network/prefix, returning them in
        # reverse order, so that we can choose the first one.
        possible_ancestors = Network.objects.filter(**lookup_kwargs).order_by(
            "-prefix_length"
        )

        # If we've got any matches, the first one is our closest parent.
        try:
            return possible_ancestors[0]
        except IndexError:
            raise Network.DoesNotExist(
                "Network matching query does not exist."
            )

    def reserved(self):
        return Network.objects.filter(state=Network.RESERVED)


class Network(Resource):
    """Represents a subnet or IP address."""

    ALLOCATED = "allocated"
    ASSIGNED = "assigned"
    ORPHANED = "orphaned"
    RESERVED = "reserved"

    STATE_CHOICES = (
        (ALLOCATED, ALLOCATED.title()),
        (ASSIGNED, ASSIGNED.title()),
        (ORPHANED, ORPHANED.title()),
        (RESERVED, RESERVED.title()),
    )
    BUSY_STATES = [ASSIGNED, RESERVED]

    network_address = fields.BinaryIPAddressField(
        max_length=16,
        null=False,
        db_index=True,
        verbose_name="Network Address",
        help_text=(
            "The network address for the Network. The network address and "
            "the prefix length together uniquely define a network."
        ),
    )
    broadcast_address = fields.BinaryIPAddressField(
        max_length=16,
        null=False,
        db_index=True,
        help_text="The broadcast address for the Network. (Internal use only)",
    )
    prefix_length = models.IntegerField(
        null=False,
        db_index=True,
        verbose_name="Prefix Length",
        help_text="Length of the Network prefix, in bits.",
    )
    ip_version = models.CharField(
        max_length=1,
        null=False,
        db_index=True,
        choices=constants.IP_VERSION_CHOICES,
    )
    is_ip = models.BooleanField(
        null=False,
        default=False,
        db_index=True,
        editable=False,
        help_text="Whether the Network is a host address or not.",
    )
    site = models.ForeignKey(
        "Site",
        db_index=True,
        related_name="networks",
        on_delete=models.PROTECT,
        verbose_name="Site",
        help_text="Unique ID of the Site this Network is under.",
    )
    parent = models.ForeignKey(
        "self",
        blank=True,
        null=True,
        related_name="children",
        default=None,
        db_index=True,
        on_delete=models.PROTECT,
        help_text="The parent Network of the Network.",
    )
    state = models.CharField(
        max_length=20,
        null=False,
        db_index=True,
        choices=STATE_CHOICES,
        default=ALLOCATED,
        help_text="The allocation state of the Network.",
    )

    # Implements .objects.get_by_address() and .get_closest_parent()
    objects = NetworkManager()

    def __init__(self, *args, **kwargs):
        self._cidr = kwargs.pop("cidr", None)
        super().__init__(*args, **kwargs)

    def __str__(self):
        return self.cidr

    class Meta:
        unique_together = (
            "site",
            "ip_version",
            "network_address",
            "prefix_length",
        )

    def supernets(self, direct=False, discover_mode=False, for_update=False):
        query = Network.objects.all()

        if self.parent is None and not discover_mode:
            return query.none()

        if discover_mode and direct:
            raise exc.ValidationError(
                "Direct is incompatible with discover_mode."
            )

        if for_update:
            query = query.select_for_update()

        if direct:
            return query.filter(id=self.parent.id)

        return query.filter(
            site=self.site,
            is_ip=False,
            ip_version=self.ip_version,
            prefix_length__lt=self.prefix_length,
            network_address__lte=self.network_address,
            broadcast_address__gte=self.broadcast_address,
        )

    def subnets(
        self,
        include_networks=True,
        include_ips=True,
        direct=False,
        for_update=False,
    ):
        query = Network.objects.all()

        if not any([include_networks, include_ips]) or self.is_ip:
            return query.none()

        if for_update:
            query = query.select_for_update()

        if not all([include_networks, include_ips]):
            if include_networks:
                query = query.filter(is_ip=False)
            if include_ips:
                query = query.filter(is_ip=True)

        if direct:
            return query.filter(parent__id=self.id)

        return query.filter(
            site=self.site,
            ip_version=self.ip_version,
            prefix_length__gt=self.prefix_length,
            network_address__gte=self.network_address,
            broadcast_address__lte=self.broadcast_address,
        )

    def get_next_network(
        self, prefix_length, num=1, strict=False, as_objects=True
    ):
        """
        Return a list of the next available networks.

        If no networks are available, an empty list will be returned.

        :param prefix_length:
            The prefix length of networks

        :param num:
            The number of networks desired

        :param as_objects:
            Whether to return IPNetwork objects or strings

        :param strict:
            Whether to return networks for strict allocation

        :returns:
            list(IPNetwork)
        """
        start_time = time.time()  # For debugging

        # If we're reserved, automatically ZILCH!!
        # TODO(jathan): Should we raise an error instead?
        if self.state == Network.RESERVED:
            return []

        try:
            prefix_length = int(prefix_length)
        except (TypeError, ValueError) as err:
            raise exc.ValidationError({"prefix_length": str(err)})

        if prefix_length < self.prefix_length:
            raise exc.ValidationError(
                {
                    "prefix_length": "New prefix must be longer than %r"
                    % self.prefix_length
                }
            )

        try:
            num = int(num)
        except (TypeError, ValueError) as err:
            raise exc.ValidationError({"num": str(err)})

        # Always default to 1 for numbers less than 1.
        if num < 1:
            num = 1

        cidr = self.ip_network

        if prefix_length > cidr.max_prefixlen:
            try:
                next(cidr.subnets(new_prefix=prefix_length))
            except ValueError as err:
                raise exc.ValidationError({"prefix_length": str(err)})

        if strict:
            children = [c.ip_network for c in self.get_children()]
        else:
            children = [
                c.ip_network
                for c in self.get_descendants()
                if (c.prefix_length >= prefix_length)
            ]

        exclude_nums = {}

        network_prefix = cidr.network_address

        # Get integer value of network address of parent network shifted
        # (cidr.max_prefixlen - prefix_length) bits to the right
        a = int(network_prefix) >> (cidr.max_prefixlen - prefix_length)

        for c in children:
            # For each child get integer value of network address shifted
            # (cidr.max_prefixlen - prefix_length) bits to the right
            b = int(c.network_address) >> (cidr.max_prefixlen - prefix_length)

            # Get xor of parent network address and child network address this
            # gets rid of the parent network address bits
            d = a ^ b

            # Store the child's prefix length in excluded_nums with the
            # variable d as the key
            if d in exclude_nums:
                # If two children share the same key, then store the shortest
                # prefix length
                if c.prefixlen < exclude_nums[d]:
                    exclude_nums[d] = c.prefixlen
            else:
                exclude_nums[d] = c.prefixlen

        wanted = []

        # Keep a counter starting at integer value of parent network address
        counter = int(cidr.network_address)

        # The upper limit is parent network prefix + 1
        upper = int(cidr.network_address) + 2 ** (
            cidr.max_prefixlen - cidr.prefixlen
        )

        while counter < upper:
            # If we have requested number of networks then we can break
            if len(wanted) == num:
                break

            if cidr.version == 4:
                next_subnet = ipaddress.IPv4Network((counter, prefix_length))
            else:
                next_subnet = ipaddress.IPv6Network((counter, prefix_length))

            # Shift the bits between parent prefix and requested prefix all the
            # way to the right
            b = counter >> (cidr.max_prefixlen - prefix_length)

            # Remove the parent network address part
            c = a ^ b

            if c in exclude_nums:
                # If this sequence of bits were seen before then we must skip
                # this network
                p = exclude_nums.pop(c)

                if p < prefix_length:
                    # If current network is possibly child of another child
                    # then we must skip overlapping child's range of addresses,
                    # this is so we can implement strict allocation
                    counter += 2 ** (cidr.max_prefixlen - p)
                else:
                    # Otherwise just skip to next network with requested
                    # prefix_length
                    counter += 2 ** (cidr.max_prefixlen - prefix_length)
                continue
            counter += 2 ** (cidr.max_prefixlen - prefix_length)

            # If this is an interconnect network, we include first and last
            # address in subnet
            if cidr.prefixlen in settings.NETWORK_INTERCONNECT_PREFIXES:
                pass
            elif prefix_length in settings.HOST_PREFIXES and (
                next_subnet.network_address == cidr.network_address
                or next_subnet.broadcast_address == cidr.broadcast_address
            ):
                # Otherwise we skip first and last address in subnet
                continue

            # Add network to wanted list
            wanted.append(next_subnet)

        elapsed_time = time.time() - start_time
        log.debug(">> WANTED = %s", wanted)
        log.debug(">> ELAPSED TIME: %s" % elapsed_time)
        return wanted if as_objects else [str(w) for w in wanted]

    def get_next_address(self, num=1, strict=False, as_objects=True):
        """
        Return a list of the next available addresses.

        If no addresses are available, an empty list will be returned.

        :param num:
            The number of addresses desired

        :param as_objects:
            Whether to return IPNetwork objects or strings
        """
        prefix_map = {"4": 32, "6": 128}  # Map ip_version => prefix_length
        prefix_length = prefix_map.get(self.ip_version)

        return self.get_next_network(
            prefix_length=prefix_length,
            num=num,
            strict=strict,
            as_objects=as_objects,
        )

    def is_child_node(self):
        """
        Returns whether I am a child node.
        """
        return self.parent is not None

    def is_leaf_node(self):
        """
        Returns whether I am leaf node (no children).
        """
        return not self.children.exists()

    def is_root_node(self):
        """
        Returns whether I am a root node (no parent).
        """
        return self.parent is None

    def get_ancestors(self, ascending=False):
        """Return my ancestors."""
        query = self.supernets().order_by("network_address", "prefix_length")
        if ascending:
            query = query.reverse()
        return query

    def get_children(self):
        """Return my immediate children."""
        return self.subnets(include_ips=True, direct=True).order_by(
            "network_address", "prefix_length"
        )

    def get_descendants(self):
        """Return all of my children!"""
        return self.subnets(include_ips=True).order_by(
            "network_address", "prefix_length"
        )

    def get_root(self):
        """
        Returns the root node (the parent of all of my ancestors).
        """
        ancestors = self.get_ancestors()
        return ancestors.first()

    def get_siblings(self, include_self=False):
        """
        Return my siblings. Root nodes are siblings to other root nodes.
        """
        query = Network.objects.filter(
            parent=self.parent, site=self.site
        ).order_by("network_address", "prefix_length")
        if not include_self:
            query = query.exclude(id=self.id)

        return query

    def get_utilization(self):
        return util.get_network_utilization(self)

    def set_reserved(self, commit=True):
        self.state = self.RESERVED
        if commit:
            self.save()

    def set_allocated(self, commit=True):
        self.state = self.ALLOCATED
        if commit:
            self.save()

    def set_assigned(self, commit=True):
        self.state = self.ASSIGNED
        if commit:
            self.save()

    def set_orphaned(self, commit=True):
        self.state = self.ORPHANED
        if commit:
            self.save()

    @property
    def cidr(self):
        return "%s/%s" % (self.network_address, self.prefix_length)

    @property
    def ip_network(self):
        return ipaddress.ip_network(self.cidr)

    def reparent_subnets(self):
        """
        Determine list of child nodes and set the parent to self.
        """
        query = Network.objects.select_for_update().filter(
            ~models.Q(id=self.id),  # Don't include yourself...
            parent_id=self.parent_id,
            prefix_length__gt=self.prefix_length,
            ip_version=self.ip_version,
            network_address__gte=self.network_address,
            broadcast_address__lte=self.broadcast_address,
        )

        query.update(parent=self)

    def clean_state(self, value):
        """Enforce that state is one of the valid states."""
        value = value.lower()
        if value not in [s[0] for s in self.STATE_CHOICES]:
            raise exc.ValidationError({"state": "Invalid state: %r" % value})

        return value

    def clean_fields(self, exclude=None):
        """This will enforce correct values on fields."""
        cidr = self._cidr
        if cidr is None and self.network_address and self.prefix_length:
            cidr = "%s/%s" % (self.network_address, self.prefix_length)

        if not cidr:
            msg = f"Invalid CIDR: {cidr}. Must be IPv4/IPv6 notation."
            raise exc.ValidationError(msg)

        # Determine network properties
        network = cidr  # In-case we're not a string.

        # Convert a string to an IPNetwork.
        if isinstance(cidr, str):
            try:
                network = ipaddress.ip_network(cidr)
            except ValueError as err:
                raise exc.ValidationError({"cidr": str(err)})

        if network.network_address == network.broadcast_address:
            self.is_ip = True

        self.ip_version = str(network.version)
        self.network_address = str(network.network_address)
        self.broadcast_address = str(network.broadcast_address)
        self.prefix_length = network.prefixlen
        self.state = self.clean_state(self.state)

    # Shoutout to jathanism for this code.
    def delete(self, **kwargs):
        force_delete = kwargs.pop("force_delete", False)

        # Block deletion if this Network has active Interface assignments.
        # force_delete does NOT bypass this — it is only for reparenting
        # child networks.
        if self.assignments.exists():
            raise exc.ProtectedError(
                "Cannot delete Network %s because it has active Interface "
                "assignments. Remove the assignments first." % self.cidr,
                set(self.assignments.all()),
            )

        try:
            super().delete(**kwargs)
        except exc.ProtectedError as err:
            if force_delete:
                new_parent = self.parent
                # Check if the network does not have a parent, check that it's
                # children are not leaf nodes. If so, raise an error.
                if not new_parent:
                    children = self.get_children()
                    for child in children:
                        if child.is_leaf_node():
                            raise exc.Conflict(
                                "You cannot forcefully delete a network that "
                                "does not have a parent, and whose children "
                                "are leaf nodes."
                            )
                # Otherwise, update all children to use the new parent and
                # delete the old parent of these child nodes.
                # In Django 4.1+, protected_objects is a set of model
                # instances rather than a QuerySet, so we always extract
                # the PKs and use an explicit QuerySet to perform the
                # update.
                protected_ids = [p.pk for p in err.protected_objects]
                Network.objects.filter(pk__in=protected_ids).update(
                    parent=new_parent
                )
                super().delete(**kwargs)
            else:
                raise

    def save(self, *args, **kwargs):
        """This is stuff we want to happen upon save."""
        self.full_clean()  # First validate fields are correct

        for_update = kwargs.pop("for_update", False)

        # Calculate our supernets and determine if we require a parent.
        supernets = self.supernets(discover_mode=True, for_update=for_update)
        if supernets:
            parent = max(supernets, key=attrgetter("prefix_length"))
            self.parent = parent

        if self.parent is None and self.is_ip:
            raise exc.ValidationError("IP Address needs base network.")

        # Save, so we get an ID, and register our parent.
        # Skip full_clean in Resource.save() since we already did it above.
        super().save(*args, _skip_full_clean=True, **kwargs)

        # If we're not an IP, determine our subnets and reparent them.
        if not self.is_ip:
            self.reparent_subnets()

    def to_dict(self):
        return {
            "id": self.id,
            "cidr": self.cidr,
            "parent_id": self.parent_id,
            "parent": self.parent and self.parent.cidr,
            "site_id": self.site_id,
            "is_ip": self.is_ip,
            "ip_version": self.ip_version,
            "network_address": self.network_address,
            "prefix_length": self.prefix_length,
            "state": self.state,
            "attributes": self.get_attributes(),
        }


# Signals
def refresh_assignment_interface_networks(sender, instance, **kwargs):
    """This signal fires each time a Network object is saved. Upon save,
    the signal iterates through all the child networks of the network
    being saved and cleans the addresses and networks assigned to the
    interfaces (if any) to which these child networks have been assigned.

    We need to clean the addresses on an Interface upon a call to save()
    on Network due to the Interface model caching _addresses & _networks
    which causes the update on the Network object to not cascade onto the
    corresponding Interface object."""
    for child in instance.children.all():
        for assignment in child.assignments.all():
            assignment.interface.clean_addresses()
            assignment.interface.save()


models.signals.post_save.connect(
    refresh_assignment_interface_networks,
    sender=Network,
    dispatch_uid="refresh_interface_assignment_networks_post_save_network",
)



# -*- coding: utf-8 -*-

from django.db import models, migrations

class Migration(migrations.Migration):

    dependencies = [
        ("nsot", "0006_auto_20150810_1947"),
    ]

    operations = [
        migrations.RenameField(
            model_name="interface",
            old_name="mac",
            new_name="mac_address",
        ),
    ]



from django.conf import settings
from django.db import models

from .. import exc
from .circuit import Circuit
from .resource import Resource


class Device(Resource):
    """Represents a network device."""

    hostname = models.CharField(
        max_length=255,
        null=False,
        db_index=True,
        help_text="The hostname of the Device.",
    )
    site = models.ForeignKey(
        "Site",
        db_index=True,
        related_name="devices",
        on_delete=models.PROTECT,
        verbose_name="Site",
        help_text="Unique ID of the Site this Device is under.",
    )

    def __str__(self):
        return "%s" % self.hostname

    class Meta:
        unique_together = ("site", "hostname")

    @property
    def circuits(self):
        """All circuits related to this Device."""
        interfaces = self.interfaces.all()
        circuits = []
        for intf in interfaces:
            try:
                circuits.append(intf.circuit)
            except Circuit.DoesNotExist:
                continue
        return circuits

    def clean_hostname(self, value):
        if not value:
            raise exc.ValidationError(
                {"hostname": "Hostname must be non-zero length string."}
            )
        if len(value) > 253:
            raise exc.ValidationError(
                {"hostname": "Hostname exceeds 253-character FQDN limit."}
            )
        if not settings.DEVICE_NAME.match(value):
            raise exc.ValidationError(
                {"hostname": "Invalid hostname: %r." % value}
            )
        return value

    def clean_fields(self, exclude=None):
        self.hostname = self.clean_hostname(self.hostname)

    def to_dict(self):
        return {
            "id": self.id,
            "site_id": self.site_id,
            "hostname": self.hostname,
            "attributes": self.get_attributes(),
        }



from rest_framework import status
from rest_framework.mixins import CreateModelMixin
from rest_framework.response import Response

__all__ = [
    "BulkCreateModelMixin",
    "BulkDestroyModelMixin",
    "BulkUpdateModelMixin",
]


class BulkCreateModelMixin(CreateModelMixin):
    """
    Either create a single or many model instances in bulk by using the
    Serializers ``many=True`` ability from Django REST >= 2.2.5.

    .. note::
        This mixin uses the same method to create model instances
        as ``CreateModelMixin`` because both non-bulk and bulk
        requests will use ``POST`` request method.
    """

    def create(self, request, *args, **kwargs):
        bulk = isinstance(request.data, list)

        if not bulk:
            return super(BulkCreateModelMixin, self).create(
                request, *args, **kwargs
            )

        else:
            serializer = self.get_serializer(data=request.data, many=True)
            serializer.is_valid(raise_exception=True)
            self.perform_bulk_create(serializer)
            return Response(serializer.data, status=status.HTTP_201_CREATED)

    def perform_bulk_create(self, serializer):
        return self.perform_create(serializer)


class BulkUpdateModelMixin(object):
    """
    Update model instances in bulk by using the Serializers
    ``many=True`` ability from Django REST >= 2.2.5.
    """

    def get_object(self):
        lookup_url_kwarg = self.lookup_url_kwarg or self.lookup_field

        if lookup_url_kwarg in self.kwargs:
            return super(BulkUpdateModelMixin, self).get_object()

        # If the lookup_url_kwarg is not present
        # get_object() is most likely called as part of options()
        # which by default simply checks for object permissions
        # and raises permission denied if necessary.
        # Here we don't need to check for general permissions
        # and can simply return None since general permissions
        # are checked in initial() which always gets executed
        # before any of the API actions (e.g. create, update, etc)
        return

    def bulk_update(self, request, *args, **kwargs):
        partial = kwargs.pop("partial", False)

        # restrict the update to the filtered queryset
        serializer = self.get_serializer(
            self.filter_queryset(self.get_queryset()),
            data=request.data,
            many=True,
            partial=partial,
        )
        serializer.is_valid(raise_exception=True)
        self.perform_bulk_update(serializer)
        return Response(serializer.data, status=status.HTTP_200_OK)

    def partial_bulk_update(self, request, *args, **kwargs):
        kwargs["partial"] = True
        return self.bulk_update(request, *args, **kwargs)

    def perform_update(self, serializer):
        serializer.save()

    def perform_bulk_update(self, serializer):
        return self.perform_update(serializer)


class BulkDestroyModelMixin(object):
    """
    Destroy model instances.
    """

    def allow_bulk_destroy(self, qs, filtered):
        """
        Hook to ensure that the bulk destroy should be allowed.

        By default this checks that the destroy is only applied to
        filtered querysets.
        """
        return qs is not filtered

    def bulk_destroy(self, request, *args, **kwargs):
        qs = self.get_queryset()

        filtered = self.filter_queryset(qs)
        if not self.allow_bulk_destroy(qs, filtered):
            return Response(status=status.HTTP_400_BAD_REQUEST)

        self.perform_bulk_destroy(filtered)

        return Response(status=status.HTTP_204_NO_CONTENT)

    def perform_destroy(self, instance):
        instance.delete()

    def perform_bulk_destroy(self, objects):
        for obj in objects:
            self.perform_destroy(obj)



"""
Middleware for authentication.
"""

import logging

from django.conf import settings
from django.contrib.auth import backends, middleware
from django.core.exceptions import ValidationError
from django.core.validators import EmailValidator
from guardian.backends import ObjectPermissionBackend
from guardian.core import ObjectPermissionChecker

from ..util import normalize_auth_header

log = logging.getLogger("nsot_server")


class EmailHeaderMiddleware(middleware.RemoteUserMiddleware):
    header = normalize_auth_header(settings.USER_AUTH_HEADER)


class EmailHeaderBackend(backends.RemoteUserBackend):
    """Custom backend that validates username is an email."""

    def authenticate(self, request, remote_user):
        """Override default to return None if username is invalid."""
        if not remote_user:
            return None

        username = self.clean_username(remote_user)
        if not username:
            return None

        return super().authenticate(request, remote_user)

    def clean_username(self, username):
        """Makes sure that the username is a valid email address."""
        validator = EmailValidator()
        try:
            validator(username)  # If invalid, will raise a ValidationError
        except ValidationError:
            log.debug("Invalid email address: %r", username)
            return None
        else:
            return username

    def configure_user(self, request, user, created=True):
        """Configure newly created users.

        In Django 5.x this is called for all users (new and existing)
        with a ``created`` flag.  Only grant superuser on first creation.
        """
        if created:
            if settings.NSOT_NEW_USERS_AS_SUPERUSER:
                user.is_superuser = True
                user.is_staff = True
                user.save()
            log.debug("Created new user: %s", user)
        return user


class NsotObjectPermissionsBackend(ObjectPermissionBackend):
    """Custom backend that overloads django-guardian's has_perm method."""

    def has_perm(self, user_obj, perm, obj=None):
        """
        Returns ``True`` if ``user_obj`` has ``perm`` for ``obj``. If no
        ``obj`` is provided, ``False`` is returned.

        However, if ``grp_obj`` does not have ``perm`` for ``obj``,
        the ancestor tree for ``obj`` is checked against. If any node in
        the ancestor tree has ``perm`` for the ``obj``, then ``True`` is
        returned, else ``False`` is returned.
        """
        check = super().has_perm(user_obj, perm, obj)
        if check:
            return True

        if hasattr(obj, "get_ancestors"):
            ancestors = obj.get_ancestors()
            ancestor_perm_check = ObjectPermissionChecker(user_obj)

            return any(
                ancestor_perm_check.has_perm(perm, a) for a in ancestors
            )

        return False



import logging

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import ObjectDoesNotExist
from rest_framework import authentication, exceptions

from ..util import normalize_auth_header

log = logging.getLogger(__name__)


class AuthTokenAuthentication(authentication.BaseAuthentication):
    def authenticate(self, request):
        log.debug("Fetching AuthToken header.")

        auth_header = authentication.get_authorization_header(request)
        auth_fields = auth_header.decode("utf-8").split()

        if not auth_fields or auth_fields[0].lower() != "authtoken":
            return None

        if len(auth_fields) == 1:
            raise exceptions.AuthenticationFailed(
                "Invalid token header. No credentials provided."
            )
        if len(auth_fields) > 2:
            raise exceptions.AuthenticationFailed(
                "Invalid token header. Token should not contain spaces."
            )

        _auth_type, data = auth_fields
        email, auth_token = data.split(":", 1)

        log.debug("     email: %r", email)
        log.debug("auth_token: %r", auth_token)

        return self.authenticate_credentials(email, auth_token)

    def authenticate_credentials(self, email, auth_token):
        user = get_user_model().verify_auth_token(email, auth_token)

        # If user is bad this time, it's an invalid login
        if user is None:
            raise exceptions.AuthenticationFailed(
                "Invalid login/token expired."
            )
            # raise exc.Unauthorized('Invalid login/token expired.')

        log.debug("token_auth authenticated user: %s" % email)

        return (user, auth_token)

    def authenticate_header(self, request):
        return "AuthToken"


class EmailHeaderAuthentication(authentication.BaseAuthentication):
    def authenticate(self, request):
        user_auth_header = normalize_auth_header(settings.USER_AUTH_HEADER)
        log.debug(
            "EmailHeaderAuthentication.authenticate(): auth_header = %r",
            user_auth_header,
        )

        # Naively fetch the user email from the auth_header
        email = request.META.get(user_auth_header)
        log.debug(
            "EmailHeaderAuthentication.authenticate(): email = %r", email
        )
        if email is None:
            return None

        # Fetch a stinkin' user
        try:
            user = get_user_model().objects.get(email=email)
        except ObjectDoesNotExist:
            # Make this a 400 for now since it's failing validation.
            raise exceptions.ValidationError(
                "Username must contain a valid email address"
            )

        # And return it.
        return (user, None)


class SecretKeyAuthentication(authentication.BaseAuthentication):
    def authenticate_credentials(self, email, secret_key):
        try:
            user = get_user_model().objects.get(email=email)
        except ObjectDoesNotExist:
            user = None

        # Make sure we've got a user and the secret_key is valid
        if user is not None and user.verify_secret_key(secret_key):
            return user, secret_key  # Auth success

        raise exceptions.AuthenticationFailed("Invalid email/secret_key")
